import numpy as np
import copy
from collections import defaultdict

from sistem.utilities.utilities import get_reg_id, update

class Chromosome:
    def __init__(self, name='', allele=None, homolog_id=None, seq=None, cent=None):
        self.name = name
        self.allele = allele
        self.homolog_id = homolog_id
        self.seq = seq
        self.cent = cent

    def __len__(self):
        return len(self.seq)
    
    def intact(self):
        return len(self.seq)

    def copy(self, homolog_id=None):
        if homolog_id == None:
            homolog_id = self.homolog_id
        new_chrom = Chromosome(name=self.name, allele=self.allele, homolog_id=homolog_id, seq=copy.copy(self.seq), cent=self.cent)
        return new_chrom
    
    # place m additional copies of the segment from start to end in tandem after m
    def amplify(self, start, end, m):
        new_seg = m * self.seq[start:end]
        self.seq = self.seq[:end] + new_seg + self.seq[end:]
        if self.cent is not None and self.cent >= end:
            self.cent += m*(end - start)
    
    # Delete the segment from start to end
    def delete(self, start, end):
        self.seq = self.seq[:start] + self.seq[end:]
        if self.cent is not None:
            if self.cent > start and self.cent < end:
                self.cent = start
            if self.cent >= end:
                self.cent -= (end - start)
            if self.cent == 0 or self.cent >= len(self.seq):
                self.cent = None



def default_SNV_obj():
    return defaultdict(int)

class SNVChromosome(Chromosome):
    """
    Attributes:
        unique_regions (defaultdict(int)): counts the number of times some region r has appeared to create unique region id
        SNVs (defaultdict(defaultdict(int))): dictionary of dictionary of ints. First key is region id; Second key is the position in that region (between 0 and region_len - 1), third is either 0,1,2, referring to the non-reference base pair.

    """
    def __init__(self, unique_regions=None, SNVs=None, driverSNV_counts=None, ndriverSNVs=None, **kwargs):
        self.unique_regions = unique_regions if unique_regions is not None else defaultdict(int)
        self.SNVs = SNVs if SNVs is not None else defaultdict(default_SNV_obj)
        #self.driverSNV_counts = driverSNV_counts if driverSNV_counts is not None else defaultdict(int)
        super().__init__(**kwargs)

    def copy(self, homolog_id=None):
        if homolog_id == None:
            homolog_id = self.homolog_id
        new_chrom = SNVChromosome(name=self.name, allele=self.allele, homolog_id=homolog_id, seq=copy.copy(self.seq), cent=self.cent, unique_regions=copy.copy(self.unique_regions), SNVs=copy.copy(self.SNVs))#, driverSNV_counts = copy.copy(self.driverSNV_counts))
        return new_chrom
    
    def up(self, r):
        v = get_reg_id(r)
        self.unique_regions[v] += 1
        return update(v, self.unique_regions[v])

    def amplify(self, start, end, m):
        gain_seg = self.seq[start:end]
        #driverSNV_regions = [i for i,r in enumerate(gain_seg) if r in self.driverSNV_counts]
        SNV_regions = [i for i,r in enumerate(gain_seg) if r in self.SNVs]
        new_seg = []
        for _ in range(m):
            temp_new_seg = [self.up(r) for r in gain_seg]
            new_seg.extend(temp_new_seg)
            #for i in driverSNV_regions:
            #    self.driverSNV_counts[temp_new_seg[i]] = np.copy(self.driverSNV_counts[gain_seg[i]])
            for i in SNV_regions:
                self.SNVs[temp_new_seg[i]] = copy.copy(self.SNVs[gain_seg[i]])
        self.seq = self.seq[:end] + new_seg + self.seq[end:]
        if self.cent is not None and self.cent >= end:
            self.cent += m*(end - start)
    
    def delete(self, start, end):
        for r in self.seq[start:end]:
            #if r in self.driverSNV_counts:
            #    del self.driverSNV_counts[r]
            if r in self.SNVs:
                del self.SNVs[r]
        super().delete(start, end)
    
    def add_SNV(self, start, position, bp):
        self.SNVs[int(self.seq[start])][position] = bp


import subprocess

from sistem.genome import SNVChromosome
from sistem.utilities.utilities import get_reg_id

def build_cell_ref(chromosome, ref, num_regions, region_len, out_path):
    nucs = ['A', 'C', 'G', 'T']
    with open(out_path, 'w+') as f:
        f.write(f'>{chromosome.name}\n')
        for r in chromosome.seq:
            q = get_reg_id(r)
            if q == num_regions - 1:
                ref_seq = list(ref[chromosome.name][int(q*region_len):].seq.upper())
            else:
                ref_seq = list(ref[chromosome.name][int(q*region_len):int((q+1)*region_len)].seq.upper())
            if isinstance(chromosome, SNVChromosome):
                if r in chromosome.SNVs:
                    positions = chromosome.SNVs[r]
                    for p,idx in positions.items():
                        if p < len(ref_seq) - 1:
                            if ref_seq[p] != 'N':
                                alt_nucs = nucs[:]
                                alt_nucs.remove(ref_seq[p])
                                alt = alt_nucs[idx]
                                ref_seq[p] = alt
            f.write("".join(ref_seq))
    call = subprocess.run(['samtools', 'faidx', out_path])



from collections import deque
import os
from typing import Optional

class Node:
    def __init__(
        self, 
        name=None, 
        parent=None, 
        edge_length=0, 
        **kwargs
    ):
        self.name = name
        self.parent = parent
        self.length = edge_length
        self.children = []
        self.info = {}
        #super().__init__(**kwargs)
    
    def __str__(self):
        if self.name:
            return str(self.name)
        else:
            return str(id(self))

    def __len__(self):
        return len(self.iter_leaves())
    
    #format codes --> 0: leaf names only, 1: leaf names + lengths only, 2: leaf and internal names, 3: leaf and internal names + lengths
    def _write_newick(self, terminate=True, format=3):
        if self.is_leaf():
            if format == 0 or format == 2:
                return str(self)
            elif format == 1 or format == 3:
                return str(self) + ':' + str(self.length)
        else:
            newick_str = '('
            for child in self.children:
                newick_str += child._write_newick(terminate=False, format=format) + ','
            newick_str = newick_str[:-1]
            newick_str += ')'
            if format == 2 or format == 3 or terminate:
                newick_str += str(self)
            if (format == 1 or format == 3) and not terminate :
                newick_str += ':' + str(self.length)
        if terminate:
            return newick_str + ';'
        else:
            return newick_str
    
    def is_leaf(self):
        return len(self.children) == 0
    
    def is_root(self):
        return self.parent is None

    def get_root(self):
        root = self
        while root.parent is not None:
            root = root.parent
        return root
    
    def add_child(self, node):
        self.children.append(node)
        node.parent = self
    
    def detach(self):
        if not self.is_root():
            self.parent.children.remove(self)
            self.parent = None

    #level order traversal, i.e. breadth first search from the root. Does include the root itself.
    def iter_descendants(self, include_root=True):
        nodes = deque()
        if include_root:
            nodes.append(self)
        else:
            nodes.extend(self.children)

        while nodes:
            node = nodes.popleft()
            nodes.extend(node.children)
            yield node
    
    def iter_preorder(self):
        visit_queue = deque()
        visit_queue.append(self)

        while visit_queue:
            node = visit_queue.pop()
            visit_queue.extend(node.children)
            yield node
    
    def iter_postorder(self):
        visit_queue = deque()
        return_queue = deque()
        visit_queue.append(self)

        while visit_queue:
            node = visit_queue.pop()
            return_queue.append(node)
            if not node.is_leaf():
                visit_queue.extend(node.children)
        
        while return_queue:
            node = return_queue.pop()
            yield node
    
    def iter_leaves(self):
        return [n for n in self.iter_postorder() if n.is_leaf()]

    def get_height(self):
        if self.is_leaf():
            return 0
        else:
            return max([1 + child.get_height() for child in self.children])

    def get_total_branchlen(self):
        return sum([node.length for node in self.iter_descendants()])
    
    def is_ancestor(self, other):
        for node in self.iter_descendants():
            if node == other:
                return True
        return False
    
class Tree:
    def __init__(
        self, 
        root: Optional[Node] = None, 
        newick: Optional[str] = None
    ):
        self.root = root
        if newick:
            if os.path.exists(newick):
                with open(newick) as f:
                    newick_str = f.readline().strip()
                    self.root = str_to_newick(newick_str)
            else:
                self.root = str_to_newick(newick)
    
    
    def __repr__(self):
        return self.root._write_newick(format=3)

    def __len__(self):
        return len([leaf for leaf in self.iter_leaves()])

    def __iter__(self):
        return self.iter_leaves()

    def print_newick(self, format=3):
        newick_str = self.root._write_newick(format=format)
        print(newick_str)
    
    def write_newick(self, format=3):
        newick_str = self.root._write_newick(format=format)
        return newick_str

    #post order traversal, i.e. left,right,middle.
    def iter_postorder(self):
        return self.root.iter_postorder()
    
    #level order traversal, i.e. breadth first search from the root. Does include the root itself.
    def iter_descendants(self, include_root=True):
        return self.root.iter_descendants(include_root=include_root)
    
    def iter_preorder(self):
        return self.root.iter_preorder()
    
    def traverse(self):
        return self.iter_descendants()
    
    def iter_leaves(self):
        for node in self.iter_descendants():
            if node.is_leaf():
                yield node
    
    def iter_ancestors(self):
        for node in self.iter_descendants():
            if not node.is_leaf():
                yield node
    
    def iter_path_to_node(self, node):
        path = []
        cur_node = node
        while not cur_node.is_root():
            path.append(cur_node)
            cur_node = cur_node.parent
        path.append(self.root)
        path.reverse()
        return path

    def has_leaf_names(self):
        for leaf in self.iter_leaves():
            if leaf.name is None:
                return False
        return True

    def set_leaf_names(self):
        count = 1
        for leaf in self.iter_leaves():
            if leaf.name is None:
                leaf.name = 'cell' + str(count)
                count += 1
    
    def set_node_names(self):
        count = 0
        for n in self.iter_postorder():
            if n.name == None:
                n.name = f'A{count}'
                count += 1

    def get_tree_height(self):
        return self.root.get_height()

    def get_total_branchlen(self):
        return self.root.get_total_branchlen()

    def find(self, node_name):
        for node in self.iter_descendants():
            if node.name == node_name:
                return node
        return None

    def get_mutations(self):
        mutations = []
        for node in self.iter_descendants():
            mutations += node.events
        return mutations

    def save(self, file_path, format=3):
        newick_str = self.root._write_newick(format=format)
        f = open(file_path, 'w+')
        f.write(newick_str)
        f.close()

    def copy_structure(self):
        ts = self._write_newick(format=3)
        new_t = Tree(newick=ts)
        return new_t
    

    # Function which resolves all unifurcating nodes in a tree by merging with their children. Updates branch lengths.
    def resolve_unifurcations(self):
        nodes = [self.root]
        while nodes:
            node = nodes.pop(0)
            children = node.children
            if len(children) == 1:
                child = children[0]
                child.detach()
                node.children = child.children
                for gc in node.children:
                    gc.parent = node

                child.children = []
                node.length += child.length
                nodes.append(node)

            else:
                for c in node.children:
                    nodes.append(c)
    
    def set_branchlengths(self):
        """
        Sets the branch lengths according to difference in birth gens.
        """
        for node in self.iter_postorder():
            if not node.is_root():
                cur_gen = node.birth_gen
                par_gen = node.parent.birth_gen
                node.length = cur_gen - par_gen

def str_to_newick(newick_str):
    split_str = newick_str[:-1].split(',')

    cur_node = None
    nodes = []

    for chunk in split_str:
        while chunk[0] == '(':
            new_node = Node()
            if cur_node:
                cur_node.add_child(new_node)
            cur_node = new_node
            chunk = chunk[1:]
        rest = chunk.split(')')
        if ':' in rest[0]:
            idx = rest[0].index(':')
            try:
                edge_len = float(rest[0][idx+1:])
            except:
                edge_len = 0
            child_node = Node(name=rest[0][:idx], edge_length=edge_len)
            cur_node.add_child(child_node)
        else:
            child_node = Node(name=rest[0])
            cur_node.add_child(child_node)
        if len(rest) > 1:
            for part in rest[1:]:
                if ':' in part:
                    idx = part.index(':')
                    cur_node.name = part[:idx]
                    try:
                        edge_len=float(part[idx+1:])
                    except:
                        edge_len=0
                    cur_node.length = edge_len
                else:
                    cur_node.name = part
                    cur_node.length = 0
                if not cur_node.is_root():
                    cur_node = cur_node.parent
    return cur_node





from sistem.lineage.cell import Cell, Clone
from sistem.lineage.tree import Node, Tree
from sistem.lineage.mutation import CNA, SNV

__all__ = [
    'Cell',
    'Clone',
    'Node',
    'Tree',
    'CNA',
    'SNV'
]


import numpy as np

class Mutation:
    def __init__(self, chrom=None, allele=None, homolog_id=None, driver=True, gen=None):
        self.chrom = chrom
        self.allele = allele
        self.homolog_id = homolog_id
        self.driver = driver
        self.gen = gen
    
    def __str__(self):
        ret_str = ''
        for attr, value in vars(self).items():
            ret_str += f'{attr}: {value}, '
        return ret_str[:-2]

class CNA(Mutation):
    def __init__(self, category=None, spawn_id=None, arm=None, start=None, end=None, copies=None, gene_ids=None, **kwargs):
        self.category = category
        self.spawn_id = spawn_id
        self.arm = arm
        self.start = start
        self.gene_ids = gene_ids
        self.end = end
        self.copies = copies
        super().__init__(**kwargs)

class SNV(Mutation):
    '''
    bp: 0, 1, or 2, indicating the index of the 3 alternate base pairs ([A,C,G,T]/ref)
    '''
    def __init__(self, index=None, region=None, position=None, bp=None, **kwargs):
        self.index = index
        self.region = region
        self.position = position
        self.bp=bp
        super().__init__(**kwargs)


def derive_focal_rate(total_rate, arm_rate, chrom_rate):
    return 1 - ((1 - total_rate)/((1-arm_rate)*(1-chrom_rate)))

def draw_num_CNA_events(ncells, focal_rate, arm_rate, chromosomal_rate, WGD_rate):
    '''
    Given focal event rate, arm, chromosomal, and WGD rates, determine the number of events across all n cells of a clone
    '''
    tot_mut_rate = 1 - (1 - focal_rate) * (1 - arm_rate) * (1 - chromosomal_rate) * (1 - WGD_rate)
    nmuts = np.random.binomial(ncells, tot_mut_rate)
    return nmuts

def draw_num_SNV_events(ncells, SNV_rate):
    if SNV_rate == 0:
        return 0
    nmuts = np.random.binomial(ncells, SNV_rate)
    return nmuts

def draw_passenger_counts(cell, focal_pass_rate, SNV_pass_rate):
    num_gen = cell.length
    #npass_CNA = max(np.random.binomial(num_gen, focal_pass_rate), 1)
    npass_CNA = np.random.binomial(num_gen, focal_pass_rate)
    npass_SNV = np.random.binomial(num_gen, SNV_pass_rate)
    return npass_CNA, npass_SNV

def distribute_events(ncells, nevents_CN, nevents_SNV):
    '''
    Given n events, distribute them randomly over ncells and return an array of cell event counts, filtered if the count is > 0
    '''
    CN_counts = np.bincount(np.random.choice(ncells, nevents_CN), minlength=ncells)
    SNV_counts = np.bincount(np.random.choice(ncells, nevents_SNV), minlength=ncells)
    counts = np.stack((CN_counts, SNV_counts), axis=1)
    counts = counts[np.sum(counts, axis=1) > 0]
    return counts

def select_CNA_events(clone, nevents, focal_rate, arm_rate, chromosomal_rate, WGD_rate, focal_gain_rate, chrom_dup_rate, length_mean, mag_mean, driver=True):
    '''
    Given there are nevents, choose which events based on the relative probabiltiies
    '''
    if nevents > 0:
        rates = np.array([focal_rate, arm_rate, chromosomal_rate, WGD_rate])
        n_focal, n_arm, n_chromosomal, n_WGD = np.random.multinomial(nevents, rates / sum(rates))
        if n_WGD:
            clone.WGD()
        for i in range(n_chromosomal):
            gen_chromosomal_event(clone, chrom_dup_rate)
        for i in range(n_arm):
            gen_arm_event(clone, chrom_dup_rate)
        for i in range(n_focal):
            gen_focal_event(clone, focal_gain_rate, length_mean, mag_mean, driver=driver)

def gen_focal_event(clone, focal_gain_rate, length_mean, mag_mean, driver=True):
    '''
    Generates a focal event by drawing a chromosome and length, selecting an appropriate start location depending on driver/passenger, then selecting amp vs deletion
    '''
    poss_starts = []
    length = max(round(np.random.exponential(length_mean)), 1)

    while length > 0:
        exclude = []
        while len(poss_starts) == 0:
            chromosome = clone.genome.get_random_chrom(exclude=exclude)
            if chromosome is None: # No available chromosome
                break
            fitted_length = min(length, len(chromosome))
            if driver:
                poss_starts = clone.get_driver_start_regions(chromosome, fitted_length)
            else:
                poss_starts = clone.get_passenger_start_regions(chromosome, fitted_length)
            exclude.append(chromosome)
        if len(poss_starts) > 0:
            break
        length -= 1
        
    if len(poss_starts) == 0:
        return

    start = np.random.choice(poss_starts)
    end = start + fitted_length
    if np.random.binomial(1, focal_gain_rate):
        mag = np.random.geometric(1 / mag_mean)
        clone.focal_amplification(chromosome, start, end, mag)
    else:
        clone.focal_deletion(chromosome, start, end)

def gen_arm_event(clone, chrom_dup_rate):
    x = clone.genome.get_random_intact_arm()
    if x is not None:
        chromosome, arm = x
        if np.random.binomial(1, chrom_dup_rate):
            clone.arm_duplication(chromosome, arm)
        else:
            clone.arm_deletion(chromosome, arm)

def gen_chromosomal_event(clone, chrom_dup_rate):
    chromosome = clone.genome.get_random_intact_chrom()
    if chromosome is not None:
        if np.random.binomial(1, chrom_dup_rate):
            clone.chromosomal_duplication(chromosome)
        else:
            clone.chromosomal_deletion(chromosome)

def select_SNV_events(clone, nevents, driver=True):
    """
    Determines a set of regions to assign independent SNVs.
    """
    exclude = []
    for _ in range(nevents):
        # Start by selecting chromosome that contains appropriate regions
        poss_starts = []
        while len(poss_starts) == 0:
            chromosome = clone.genome.get_random_chrom(exclude=exclude)
            if chromosome is None: # No available chromosome
                break
            if driver:
                poss_starts = clone.get_driver_start_regions(chromosome, 1)
            else:
                poss_starts = clone.get_passenger_start_regions(chromosome, 1)
            if len(poss_starts) == 0:
                exclude.append(chromosome)

        if chromosome is None or len(poss_starts) == 0:
            return
        
        # Now choose the region and generate the event
        start = np.random.choice(poss_starts)
        gen_SNV_event(clone, chromosome, start)

def gen_SNV_event(clone, chromosome, start):
    pos = np.random.randint(clone.library.region_len)
    bp = np.random.randint(3)
    clone.add_SNV(chromosome, start, pos, bp)


__version__ = '1.0.4'

from sistem.utilities.IO import load_gs
from sistem.ancestry import GrowthSimulator
from sistem.parameters import Parameters

__all__ = [
    'load_gs',
    'GrowthSimulator',
    'Parameters',
]


from sistem.anatomy.base_anatomy import BaseAnatomy, SimpleAnatomy
from sistem.anatomy.static_anatomy import StaticAnatomy
from sistem.anatomy.genotype_anatomy import GenotypeAnatomy

__all__ = [
    'BaseAnatomy',
    'SimpleAnatomy',
    'StaticAnatomy',
    'GenotypeAnatomy'
]


from pyfaidx import Fasta
import copy

from sistem.genome.chromosome import SNVChromosome
from sistem.genome.genome import Genome

def ConvertToSNVGenome(igenome):
    '''
    Creates an identical genome made of SNVChromosome objects from an initial genome made of standard Chromosome objects

    Args:
        igenome (Genome): Starting genome made of Chromosome objects.

    Returns:
        altgenome (Genome): New genome made of SNVChromosome objects.
    '''
    altgenome = {chrname: [] for chrname in igenome}
    for chrname in altgenome:
        for chromosome in igenome[chrname]:
            altgenome[chrname].append(SNVChromosome(name=chrname, allele=chromosome.allele, homolog_id=chromosome.homolog_id, cent=chromosome.cent, seq=copy.copy(chromosome.seq)))
    g = Genome(altgenome)
    return g

# Gets the number of regions per chrom
def get_num_regions(chrom_lens, region_len):
    regions = {}
    for chrname in chrom_lens:
        nregions = int(chrom_lens[chrname] // region_len)
        rem = nregions % region_len
        if rem > 0:
            nregions += 1
        regions[chrname] = nregions
    return regions

def fixed_chrom_lens(nchroms, chrom_len):
    return {f'chr{i}': chrom_len for i in range(1, nchroms + 1)}

def hg38_chrom_lengths_from_cytoband(include_allosomes=False):
    chrom_lens = {'chr1': 249250621, 'chr10': 135534747, 'chr11': 135006516, 'chr12': 133851895, 'chr13': 115169878, 'chr14': 107349540, 'chr15': 102531392, 'chr16': 90354753, 'chr17': 81195210, 'chr18': 78077248, 'chr19': 59128983, 'chr2': 243199373, 'chr20': 63025520, 'chr21': 48129895, 'chr22': 51304566, 'chr3': 198022430, 'chr4': 191154276, 'chr5': 180915260, 'chr6': 171115067, 'chr7': 159138663, 'chr8': 146364022, 'chr9': 141213431, 'chrX': 155270560, 'chrY': 59373566}
    arm_ratios = {'chr1': 0.5015, 'chr10': 0.2966, 'chr11': 0.39776, 'chr12': 0.26746, 'chr13': 0.15542, 'chr14': 0.16395, 'chr15': 0.18531, 'chr16': 0.40507, 'chr17': 0.29558, 'chr18': 0.22029, 'chr19': 0.44817, 'chr2': 0.38364, 'chr20': 0.43633, 'chr21': 0.27426, 'chr22': 0.28652, 'chr3': 0.45954, 'chr4': 0.26366, 'chr5': 0.26753, 'chr6': 0.35649, 'chr7': 0.3764, 'chr8': 0.31155, 'chr9': 0.34699, 'chrX': 0.39029, 'chrY': 0.21053}
    if not include_allosomes:
        del chrom_lens['chrX']
        del chrom_lens['chrY']
        del arm_ratios['chrX']
        del arm_ratios['chrY']
    return chrom_lens, arm_ratios

def get_chrom_lens_from_reference(input_fasta, chrom_names=None):
    ref = Fasta(input_fasta, one_based_attributes=False)
    chrom_lens = {}
    for chrom in ref.keys():
        if chrom_names is None:
            chrom_lens[chrom] = len(ref[chrom])
        else:
            if chrom in chrom_names:
                chrom_lens[chrom] = len(ref[chrom])
    return chrom_lens
    


import copy

from sistem.selection import Attributes
from sistem.selection.arm_library import BaseArmLibrary
from sistem.genome.utils import ConvertToSNVGenome
from sistem.lineage.mutation import CNA, SNV
import sistem.lineage.mutation as mut
from sistem.utilities.utilities import combine_dicts, get_reg_id

def evolve_tree_with_passengers(tree, igenome, focal_pass_rate, SNV_pass_rate, focal_gain_rate, length_mean, mag_mean, no_leaves=False):
    """
    Master function for re-evolving a tree with passenger mutations. Tree must be resolved with appropriate branch lengths.

    Args:
        tree (Tree): A tree object made of Cell objects.
        igenome (Genome): Initial starting genome
        focal_gain_rate (float): The probability of a passenger CNA being an amplification (deletion rate is 1-focal_gain_rate).
        length_mean (int): The mean length of passenger CNAs in terms of number of regions.
        mag_mean (float): The mean number of additional copies gained during amplifications.
        no_leaves (Bool): If toggled, does not draw passenger mutations for leaf nodes
    """
    if tree.root.library.is_driver_SNV_model:
        tree.root.genome = copy.deepcopy(igenome)
    else:
        altgenome = ConvertToSNVGenome(igenome)
        tree.root.genome = copy.deepcopy(altgenome)

    reevolve_tree(tree)

    if not tree.root.library.is_driver_region_model and isinstance(tree.root.library, BaseArmLibrary):
        get_blacklisted_regions(tree.root)

    for cell in tree.iter_descendants(include_root=False):
        if no_leaves and cell.is_leaf():
            cell.inherit()
        else:
            inherit_with_passengers(cell)
            npass_CNA, npass_SNV = mut.draw_passenger_counts(cell, focal_pass_rate, SNV_pass_rate)
            for _ in range(npass_CNA):
                mut.gen_focal_event(cell, focal_gain_rate, length_mean, mag_mean, driver=False)
            if npass_SNV > 0:
                mut.select_SNV_events(cell, npass_SNV, driver=False)

def reevolve_tree(tree):
    """
    Re-applies driver mutations across tree, recording mutated subsequence of focal events along the way

    Args:
        tree (Tree): A tree object made of Cell objects.
    """
    for node in tree.iter_descendants():
        if not node.is_root():
            node.inherit()
        for event in node.events:
            apply_event(node, event)

def apply_event(cell, event):
    if isinstance(event, CNA):
        if event.category == 'WGD':
            cell.WGD(record_event=False)
        else:
            cur_chromosome = cell.genome.find(event.chrom, event.homolog_id)
            if event.category == 'focal_amplification':
                event.gene_ids = cur_chromosome.seq[event.start:event.end]
                cell.focal_amplification(cur_chromosome, event.start, event.end, event.copies, record_event=False)
            elif event.category == 'focal_deletion':
                event.gene_ids = cur_chromosome.seq[event.start:event.end]
                cell.focal_deletion(cur_chromosome, event.start, event.end, record_event=False)
            elif event.category == 'arm_duplication':
                cell.arm_duplication(cur_chromosome, event.arm, record_event=False)
            elif event.category == 'arm_deletion':
                cell.arm_deletion(cur_chromosome, event.arm, record_event=False)
            elif event.category == 'chromosomal_duplication':
                cell.chromosomal_duplication(cur_chromosome, record_event=False)
            elif event.category == 'chromosomal_deletion':
                cell.chromosomal_deletion(cur_chromosome, record_event=False)
    elif isinstance(event, SNV):
        cur_chromosome = cell.genome.find(event.chrom, event.homolog_id)
        event.region = cur_chromosome.seq[event.index]
        cell.add_SNV(cur_chromosome, event.index, event.position, event.bp, record_event=False)

def get_blacklisted_regions(node):
    """
    Function for arm library model. Finds regions that are mutated with focal driver events in the agent-based model to ensure that passenger mutations do not interfere with these events. Sets as a 'blacklisted_regions' attribute.

    Args:
        node (Node): Clone/Cell/Node object
    
    Returns:
        blacklisted_regions (dict): Dictionary of blacklisted regions.
    """
    child_regions = []
    for child in node.children:
        child_regions.append(get_blacklisted_regions(child))
    blacklisted_regions = combine_dicts(*child_regions)

    for event in node.events:
        if isinstance(event, CNA):
            if event.category == 'focal_amplification' or event.category == 'focal_deletion':
                if event.chrom not in blacklisted_regions:
                    blacklisted_regions[event.chrom] = {}
                if event.homolog_id not in blacklisted_regions[event.chrom]:
                    blacklisted_regions[event.chrom][event.homolog_id] = set()
                blacklisted_regions[event.chrom][event.homolog_id].update([get_reg_id(r) for r in event.gene_ids])
            elif event.category == 'chromosomal_duplication' or event.category == 'arm_duplication':
                if event.chrom in blacklisted_regions:
                    if event.spawn_id in blacklisted_regions[event.chrom]:
                        if event.homolog_id in blacklisted_regions[event.chrom]:
                            blacklisted_regions[event.chrom][event.homolog_id].update(blacklisted_regions[event.chrom][event.spawn_id])
                        else:
                            blacklisted_regions[event.chrom][event.homolog_id] = blacklisted_regions[event.chrom][event.spawn_id].copy()
            elif event.category == 'WGD':
                for chrname in node.genome.chrom_names:
                    if chrname in blacklisted_regions:
                        n = int(len(node.genome[chrname])/2)
                        for spawn_id in range(n, len(node.genome[chrname])):
                            if spawn_id in blacklisted_regions[chrname]:
                                homolog_id = spawn_id - n
                                if homolog_id in blacklisted_regions[chrname]:
                                    blacklisted_regions[chrname][homolog_id].update(blacklisted_regions[chrname][spawn_id])
                                else:
                                    blacklisted_regions[chrname][homolog_id] = blacklisted_regions[chrname][spawn_id].copy()

    node.attributes = Attributes(blacklisted_regions=blacklisted_regions)
    return blacklisted_regions


def inherit_with_passengers(cell):
    """
    Inherits genome from parent while correcting driver mutations indices based on new passenger mutations.

    Args:
        cell (Cell): The cell to inherit.
    """
    assert cell.parent is not None and cell.parent.genome is not None, "Cannot inherit if parent and/or parent genome non-existent"
    cell.inherit()
    # If there are driver events, there is always only one
    for event in cell.events:
        if isinstance(event, CNA):
            if event.category == 'WGD':
                cell.WGD(record_event=False)
            else:
                cur_chromosome = cell.genome.find(event.chrom, event.homolog_id)
                if event.category == 'chromosomal_duplication':
                    cell.chromosomal_duplication(cur_chromosome, record_event=False)
                elif event.category == 'arm_duplication':
                    cell.arm_duplication(cur_chromosome, event.arm, record_event=False)
                elif event.category == 'chromosomal_deletion':
                    cell.chromosomal_deletion(cur_chromosome, record_event=False)
                elif event.category == 'arm_deletion':
                    cell.arm_deletion(cur_chromosome, event.arm, record_event=False)
                elif event.category == 'focal_amplification' or event.category == 'focal_deletion':
                    inherit_with_focal(event, cur_chromosome, cell.library)
        elif isinstance(event, SNV):
            cur_chromosome = cell.genome.find(event.chrom, event.homolog_id)
            inherit_with_SNV(event, cur_chromosome, cell.library)

def inherit_with_focal(event, chromosome, library):
    """
    Inherits chromosome from parent and applies driver event with updated region indices. 
    """

    # Logic for if the library uses driver regions
    if library.is_driver_region_model:
        # get number of regions that are deleted before and after the first/last driver region in mutated segment
        s1, r1, s2, r2 = get_first_last_driver(event.gene_ids, library, event.chrom)
        
        # find regions in parent chromosome, make sure regions on either side are still passengers
        #nchromosome = source_chromosome.copy()
        i1 = chromosome.seq.index(r1)
        i2 = chromosome.seq.index(r2) + 1
        for _ in range(s1):
            if i1-1 == 0 or chromosome.seq[i1 - 1] in library.drivers[event.chrom]:
                break
            i1 -= 1
        for _ in range(s2):
            if i2 >= len(chromosome.seq) or chromosome.seq[i2 + 1] in library.drivers[event.chrom]:
                break
            i2 += 1
    
    #if the library does not use driver regions (e.g. arm model), then we preserve the entire mutated subsequence
    else:
        r1,r2 = event.gene_ids[0], event.gene_ids[-1]
        i1 = chromosome.seq.index(r1)
        i2 = chromosome.seq.index(r2) + 1

    # update event with new indices
    event.start = i1
    event.end = i2
    event.gene_ids = chromosome.seq[event.start:event.end]
    if event.category == 'focal_amplification':
        chromosome.amplify(i1, i2, event.copies)
    elif event.category == 'focal_deletion':
        chromosome.delete(i1, i2)

def get_first_last_driver(segment, library, chrname):
    """
    Get the number of continous passenger regions at the beginning and end of a segment. Only used in the driver gene model.

    Args:
        segment (list): A subset of a chromosome.seq corresponding to the mutated segment.
        library (Library): Cell Library.
        chrname (str): Name of the mutated chromosome.

    Returns:
        s1 (int): Number of regions before the first driver region in the mutated sequence.
        r1 (int): Region id at s1.
        s2 (int): Number of regions before the last driver region in the mutated sequence.
        r2 (int): Region id at s2.
    """
    for i,r in enumerate(segment):
        if get_reg_id(r) in library.drivers[chrname] + library.essential[chrname]:
            s1 = i
            r1 = r
            break
    #for i,r in reversed(zip(enumerate(segment))):
    for i,r in enumerate(reversed(segment)):
        if get_reg_id(r) in library.drivers[chrname] + library.essential[chrname]:
            s2 = i
            r2 = r
            break
    return s1, r1, s2, r2


def inherit_with_SNV(event, chromosome, library):
    if library.is_driver_region_model and library.is_driver_SNV_model:
        index = chromosome.seq.index(event.region)
        event.index = index
        chromosome.add_SNV(index, event.position, event.bp)


from sistem.genome.genome import Genome
from sistem.genome.chromosome import Chromosome, SNVChromosome
from sistem.genome.utils import hg38_chrom_lengths_from_cytoband, get_chrom_lens_from_reference

__all__ = [
    'Genome',
    'Chromosome',
    'SNVChromosome',
    'hg38_chrom_lengths_from_cytoband',
    'get_chrom_lens_from_reference'
]


import string
from itertools import product

from sistem.lineage.cell import Cell

def get_ancestors(current):
    """
    Get all clones ancestral to the current clones.

    Args:
        current (defaultdict): The current/observed clones at the time of sampling as keys with the number of sampled cells as values.

    Returns:
        list: All ancestral clones.
    """
    ancestors = set()
    for clone in current:
        par = clone.parent
        while par != None:
            ancestors.add(par)
            par = par.parent
    ancestors = list(ancestors)
    return ancestors

# Assigns names to each clone in the tree with site-specific letter code and numerical clone id
def name_clone_tree(tree, nsites):
    """
    Adds names to each clone in the tree with site-specific letter code and numerical clone id.

    Args:
        tree (Tree): The clone tree.
        nsites (int): The number of anatomical sites.

    """
    chars = list(string.ascii_uppercase)
    r = 2
    while len(chars) < nsites:
        chars = [''.join(x) for x in product(chars, repeat=r)]
        r += 1
    site_counts = {i: 0 for i in range(nsites)}
    for node in tree.iter_descendants():
        if 'diploid' not in node.name:
            node.name = chars[node.site] + '-' + str(site_counts[node.site])
            site_counts[node.site] += 1

# Assigns names to new nodes in cell tree based on those already present in clone tree
def name_cell_tree(tree):
    name_dir = {}
    for node in tree.iter_descendants():
        if node.name is None:
            par_name = node.parent.name
            if '_' in par_name:
                key = par_name.split('_')[0]
            else:
                key = par_name
            if key not in name_dir:
                name_dir[key] = 1
            node.name = key + '_' + str(name_dir[key])
            name_dir[key] += 1

def add_dummy_diploids(tree, term_gen, num_diploid=1):
    n = Cell(name='diploid', birth_gen=term_gen-1, library=tree.root.library)
    tree.root.add_child(n)
    n.inherit()
    if num_diploid > 1:
        for i in range(num_diploid):
            new_n = Cell(name=f'diploid_{i+1}', birth_gen=term_gen, library=tree.root.library)
            n.add_child(new_n)
            new_n.inherit()



import numpy as np
from sklearn.preprocessing import MinMaxScaler
from pyfaidx import Fasta
import pickle as pkl
import os
import logging

from sistem.utilities.utilities import init_mbit

def load_gs(gs_path):
    if not os.path.exists(gs_path):
        raise ValueError("No file exists at the provided path")
    with open(gs_path, 'rb') as f:
        gs = pkl.load(f)
    max_region_id = max(gs.anatomy.libraries[0].regions.values()) 
    init_mbit(max_region_id.bit_length())
    return gs

def load_distances_from_file(nsites, path):
    with open(path) as f:
        lines = [l.strip().split() for l in f.readlines()]
        if len(lines) == 1:
            dists = np.array(lines[0], dtype=float).reshape(-1, 1)
            norm_dists = MinMaxScaler().fit_transform(dists)
            return norm_dists
        else:
            idxs = np.triu_indices(nsites)
            dists = np.array(lines, dtype=float)[idxs]
            norm_dists = MinMaxScaler().fit_transform(dists)
            return norm_dists

def read_fasta(input_fasta):
    ref = Fasta(input_fasta, one_based_attributes=False)
    return ref

def setup_logger(file_path=None):
    if file_path:
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        filehandler = logging.FileHandler(file_path, 'w')
        filehandler.setFormatter(logging.Formatter('%(message)s'))
        for hdlr in logger.handlers[:]:
            if isinstance(hdlr,logging.FileHandler):
                logger.removeHandler(hdlr)
        logger.addHandler(filehandler)
    else:
        logging.basicConfig(level=logging.INFO, format='%(message)s')


from collections import defaultdict
from dataclasses import dataclass, field
from typing import Optional, Union, Dict
from abc import ABC, abstractmethod

from sistem.genome.utils import get_num_regions, fixed_chrom_lens
from sistem.parameters import Parameters, fill_params
from sistem.utilities.utilities import init_mbit, get_reg_id

@dataclass
class Attributes:
    driver_counts: Optional[Dict] = field(default=None)
    arm_counts: Optional[Dict] = field(default=None)
    SNV_counts: Optional[Dict] = field(default=None)
    blacklisted_regions: Optional[Dict] = field(default=None)

class BaseLibrary(ABC):
    """Abstract base class for selection libraries. If creating a custom selection library, users must create a child class which inherits from this class. See :ref:`Parameters <parameters>` for an explanation of the constructor parameters.

    Args:
        params (Parameters, optional):
        chrom_lens (dict, optional):
        arm_ratios (float, dict, optional):
        region_len (int, optional):
        max_distinct_driv_ratio (float, optional):
        max_region_CN (int, optional):
        max_region_SNV (int, optional):
        max_ploidy (int, float, optional):
        min_ploidy (int, float, optional):

    Attributes:
        is_driver_region_model (bool): True if individual regions can be drivers, False otherwise.
        is_driver_SNV_model (bool): True if individual SNVs can impact fitness, False otherwise.

    """
    is_driver_region_model = None
    is_driver_SNV_model = None

    def __init__(
        self, 
        params: Optional[Parameters] = None,
        chrom_lens: Optional[dict] = None, 
        arm_ratios: Optional[Union[float, dict]] = None, 
        region_len: Optional[int] = None, 
        max_distinct_driv_ratio: Optional[float] = None, 
        max_region_CN: Optional[int] = None, 
        max_region_SNV: Optional[int] = None, 
        max_ploidy: Optional[Union[int, float]] = None, 
        min_ploidy: Optional[Union[int, float]] = None, 
    ):
        params = fill_params(params, chrom_lens=chrom_lens, arm_ratios=arm_ratios, region_len=region_len, max_distinct_driv_ratio=max_distinct_driv_ratio, max_region_CN=max_region_CN, max_region_SNV=max_region_SNV, max_ploidy=max_ploidy, min_ploidy=min_ploidy)
        self.chrom_lens = params.chrom_lens
        self.region_len = params.region_len
        self.arm_ratios = params.arm_ratios
        self.regions = get_num_regions(self.chrom_lens, self.region_len)

        self.max_distinct_driv_ratio = params.max_distinct_driv_ratio
        self.max_distinct_driv = 0
        self.max_region_CN = params.max_region_CN
        self.max_region_SNV = params.max_region_SNV
        self.max_ploidy = params.max_ploidy
        self.min_ploidy = params.min_ploidy

        self.delta = {chrname: [] for chrname in self.regions.keys()}
        self.base_fit = 0
        self.max_fit = 0
        
        self._ndiploid_regions = sum(list(self.regions.values()))
        max_region_id = max(self.regions.values()) 
        self._max_region_id = max_region_id
        init_mbit(max_region_id.bit_length())

    def __getitem__(self, chrname):
        return self.delta[chrname]

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # Enforce that `is_driver_region_model` is defined and valid
        if not hasattr(cls, 'is_driver_region_model') or not isinstance(cls.is_driver_region_model, bool):
            raise ValueError(
                f"Class {cls.__name__} must define a 'is_driver_region_model' attribute set to True or False, depending on whether or not the model involves driver regions."
            )

        # Enforce that `is_driver_SNV_model` is defined and valid
        if not hasattr(cls, 'is_driver_SNV_model') or not isinstance(cls.is_driver_SNV_model, bool):
            raise ValueError(
                f"Class {cls.__name__} must define a 'is_driver_SNV_model' attribute set to True or False, depending on whether or not the model involves driver SNVs."
            )

    @abstractmethod
    def compute_fitness(self, clone):
        """Computes the fitness of a clone.
        
        """

    @abstractmethod
    def init_base_fit(self):
        """Sets the base_fit attribute to the fitness of an unmutated Cell.
        """
        pass

    @abstractmethod
    def init_max_fit(self):
        """Sets the max_fit attribute to an estimate of the maximum possible fitness.
        
        """
        pass

    @abstractmethod
    def check_viability(self, clone):
        """Checks if clone passes viability checkpoints based on mutated driver stats. 

        Args:
            clone (Clone): The clone object in question.

        Returns:
            (bool): True if passes, False if not.
        
        """
        pass

    @abstractmethod
    def update_stats(self, clone, chromosome, start, end, mag=1):
        """Function which, upon a mutation occurring, updates intermediate stats stores in the clones Attributes object to make fitness computation more efficient.

        Arguments:
            clone (Clone): The clone undergoing a mutation.
            chromosome (Chromosome): The Chromosome object being mutated.
            start (int): The index of the starting region.
            end (int): The index of the ending region.
            mag (int): If the event is a CNA (start != end), then it is the number of copies being gained. If the event is an SNV (start == end), then it interprets an SNV being added in region start.
        """
        pass

    @abstractmethod
    def init_attributes(self, clone):
        """Creates an Attributes object for a clone and updates it to reflect its genome state.

        Args:
            clone (Clone): The clone object in question.
        """

    @abstractmethod
    def get_driver_start_regions(self, cell, chromosome, size):
        """Identifies the possible start region indices of a driver event.

        Args:
            cell (Cell): The current cell.
            chromosome (Chromosome): The current chromosome.
            size (int): The size in number of regions of the event (1 for SNVs, >=1 for CNAs).
        """

    @abstractmethod
    def get_passenger_start_regions(self, cell, chromosome, size):
        """Identifies the possible start region indices of a passenger event.

        Args:
            cell (Cell): The current cell.
            chromosome (Chromosome): The current chromosome.
            size (int): The size in number of regions of the event (1 for SNVs, >=1 for CNAs).
        """

    @abstractmethod
    def initialize(self, **kwargs):
        """Function used to initialize the selection coefficients.
        
        """
        pass


import numpy as np
from collections.abc import Mapping

from sistem.genome.chromosome import Chromosome, SNVChromosome

def init_diploid_genome(library):
    genome = {chrname: [] for chrname in library.regions.keys()}
    for chrname,num_regions in library.regions.items():
        if isinstance(library.arm_ratios, dict):
            arm_ratio = library.arm_ratios[chrname]
        else:
            arm_ratio = library.arm_ratios

        cent_idx = round(num_regions*arm_ratio)
        hap_profile = [i for i in range(num_regions)]
        # If SNVs are involved
        if library.is_driver_SNV_model:
            genome[chrname].append(SNVChromosome(name=chrname, allele=0, homolog_id=0, seq=[i for i in hap_profile], cent = cent_idx))
            genome[chrname].append(SNVChromosome(name=chrname, allele=1, homolog_id=1, seq=[i for i in hap_profile], cent = cent_idx))
        # If SNVs are not involved
        else:
            genome[chrname].append(Chromosome(name=chrname, allele=0, homolog_id=0, seq=[i for i in hap_profile], cent = cent_idx))
            genome[chrname].append(Chromosome(name=chrname, allele=1, homolog_id=1, seq=[i for i in hap_profile], cent = cent_idx))
    g = Genome(genome)
    return g

class Genome(Mapping):
    def __init__(self, genome, nregions=None):
        self.genome = genome
        self.nWGD = 0

        self.nregions = nregions
        if self.nregions == None:
            self.nregions = len(self)

    def __getitem__(self, chrname):
        '''
        Returns the two allele lists of the chromosome.
        '''
        return self.genome[chrname]
    
    def __iter__(self):
        '''
        Returns an iterable over chromosome names.
        '''
        return iter(self.genome)
    
    def __len__(self):
        l = 0
        for chrname in self:
            for chromosome in self[chrname]:
                l += len(chromosome)
        return l
    
    @property
    def chrom_names(self):
        return [l for l in self]
    
    def find(self, chrname, homolog_id):
        for chromosome in self[chrname]:
            if chromosome.homolog_id == homolog_id:
                return chromosome
    
    def get_chromosomes(self):
        flat_chroms = [chromosome for chrname in self for chromosome in self[chrname]]
        return flat_chroms

    def get_random_chrom(self, exclude=[]):
        '''
        Returns a random chromosome with probability proportional to its length, excluding those passed by argument.

        Args:
            exclude (list): List of Chromosome instances to exclude from selection
        
        Returns:
            chromosome (Optional[Chromosome, None]): A randomly selected chromosome if one is available, otherwise None.
        '''
        flat_chroms = [chromosome for chrname in self for chromosome in self[chrname] if chromosome not in exclude]
        chrom_sizes = [len(chromosome) for chromosome in flat_chroms]
        l = sum(chrom_sizes)
        if len(flat_chroms) == 0 or l == 0:
            return None
        p = [len(chromosome) / l for chromosome in flat_chroms]
        chromosome = np.random.choice(flat_chroms, p=p)
        return chromosome

    def get_random_intact_chrom(self):
        '''
        Returns a random chromosome which still has both arms intact
        '''
        intact_chroms = []
        for chrname in self:
            for chromosome in self[chrname]:
                if chromosome.intact() and chromosome.cent:
                    intact_chroms.append(chromosome)
        chromosome = np.random.choice(intact_chroms)
        return chromosome

    def get_random_intact_arm(self):
        '''
        Returns a chromosome arm from among all fully intact or partially intact chromosomes
        '''
        intact_arms = []
        for chrname in self:
            for chromosome in self[chrname]:
                if chromosome.intact():
                    if chromosome.cent:
                        intact_arms.append((chromosome, 'p'))
                        intact_arms.append((chromosome, 'q'))
                    else:
                        intact_arms.append((chromosome, 'w'))
        if len(intact_arms) > 0:
            idx = np.random.choice(list(range(len(intact_arms))))
            chromosome, arm = intact_arms[idx]
            return (chromosome, arm)
        else:
            return None


import numpy as np
import os
from typing import Optional, Dict

from sistem.selection.base_library import Attributes
from sistem.selection.region_library import BaseRegionLibrary
from sistem.utilities.utilities import get_reg_id
from sistem.parameters import Parameters, fill_params

class BaseHybridLibrary(BaseRegionLibrary):
    is_driver_region_model = True
    is_driver_SNV_model = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.lam = {chrname: [] for chrname in self.regions.keys()}
        
    # Needed
    def compute_fitness(self, clone):
        fitness = 1
        ploidy = clone.get_ploidy()
        for chrname in self.regions:
            for r,n in clone.attributes.driver_counts[chrname].items():
                if n == 0:
                    y = 0
                else:
                    y = clone.attributes.SNV_counts[chrname][r]/n
                k = np.exp(-y * self.lam[chrname][r])
                if r in self.drivers[chrname]:
                    s = self[chrname][r]
                    fitness *= (1 + (s*k))**(n/ploidy)
                elif r in self.essential[chrname]:
                    fitness *= k
        return fitness
    
    def check_viability(self, clone):
        p = 2**(clone.genome.nWGD + 1)
        driv_counts, SNV_counts = [], []
        for chrname in self.regions:
            for r,c in clone.attributes.driver_counts[chrname].items():
                if c != p:
                    driv_counts.append(c)
                if c != 0:
                    SNV_counts.append(clone.attributes.SNV_counts[chrname][r]/c)

        nmutated_drivers = len(driv_counts)
        if len(driv_counts) == 0:
            max_CN = p
        else:
            max_CN = max(driv_counts)
        if len(SNV_counts) == 0:
            max_SNV = 0
        else:
            max_SNV = max(SNV_counts)

        ploidy = clone.get_ploidy()

        if nmutated_drivers > self.max_distinct_driv:
            return False
        elif ploidy > self.max_ploidy:
            return False
        elif ploidy < self.min_ploidy:
            return False
        elif max_CN > self.max_region_CN:
            return False
        elif max_SNV >= self.max_region_SNV:
            return False
        else:
            return True

    def update_stats(self, clone, chromosome, start, end, mag=1):
        if start == end:
            v = get_reg_id(start)
            if v in self.drivers[chromosome.name] or v in self.essential[chromosome.name]:
                clone.attributes.SNV_counts[chromosome.name][v] += mag
        for r in chromosome.seq:
            v = get_reg_id(r)
            if v in self.drivers[chromosome.name] or v in self.essential[chromosome.name]:
                clone.attributes.driver_counts[chromosome.name][v] += mag
                clone.attributes.SNV_counts[chromosome.name][v] += mag*len(chromosome.SNVs[r])

    # Needed
    def init_attributes(self, clone):
        driver_counts = {chrname: {r: 0 for r in self.drivers[chrname] + self.essential[chrname]} for chrname in self.regions}
        SNV_counts = {chrname: {r: 0 for r in self.drivers[chrname] + self.essential[chrname]} for chrname in self.regions}
        if clone.genome is not None:
            for chrname in self.regions:
                for chromosome in clone.genome[chrname]:
                    for r in chromosome.seq:
                        v = get_reg_id(r)
                        if v in self.drivers[chrname] + self.essential[chrname]:
                            driver_counts[chrname][v] += 1
                            if r in chromosome.SNVs:
                                SNV_counts[chrname][v] += len(chromosome.SNVs[r])
        clone.attributes = Attributes(driver_counts=driver_counts, SNV_counts=SNV_counts)


class RandomHybridLibrary(BaseHybridLibrary):
    """The hybrid selection model with randomly generated selection coefficients.

    The hybrid model extends the region model to consider SNVs as well. The idea is that SNVs can disrupt the function of genes, thereby altering the selective effect of that gene on the cell's fitness. Here we distinguish between driver SNVs, which have this effect, and passenger SNVs, which have no effect (synonymous). For each region on each chromosome :math:`(k,i)`, randomly assign a second coefficient :math:`\\lambda_{k,i}` in the range (-SNV_coeff, SNV_coeff), where :math:`\\lambda_{k,i} > 0` if :math:`(k,i)` is a TSG but :math:`\\lambda_{k,i} > 0` or :math:`\\lambda_{k,i} < 0` if :math:`(k,i)` is an OG. Additionally, a fraction of NEU regions are set to be essential genes (EG), where if :math:`(k,i)` is an EG then :math:`\\delta_{k,i} = 0` and :math:`\\lambda_{k,i} > 0`, while :math:`\\lambda_{k,i} = 0` for the remaining neutral genes. Fitness is computed as

    .. math::
    
       s_a(c) = \\prod_{k=1}^K \\prod_{i=1}^{m_k} \\Big(1 + \\delta_{k,i}\\cdot(e^{-\\hat{y}_{k,i} \\lambda_{k,i}})\\Big)^{x_{k,i}/p_c} \\cdot \\Big(e^{-\\hat{y}_{k,i} \\lambda_{k,i}}\\Big)^{h(k,i)},

    where :math:`\\hat{y}_{k,i}` is the average number of driver SNVs on any copy of region :math:`i' of chromosome :math:`k' and :math:`h` is an indicator function with :math:`h(k,i) = 1` if region :math:`(k,i)` is an EG and :math:`h(k,i) = 0` otherwise.

    """
    def initialize(
        self, 
        params: Optional[Parameters] = None, 
        CN_coeff: Optional[float] = None,
        SNV_coeff: Optional[float] = None,
        OG_r: Optional[float] = None, 
        TSG_r: Optional[float] = None, 
        EG_r: Optional[float] = None
    ):
        """Method used to initialize selection coefficients. See :ref:`Parameters <parameters>` for an explanation of the parameters.

        Args:
            params (Parameters, optional):
            CN_coeff (float, optional):
            SNV_coeff (float, optional):
            OG_r (float, optional):
            TSG_r (float, optional):
            EG_r (float, optional):
        """
        params = fill_params(params, CN_coeff=CN_coeff, SNV_coeff=SNV_coeff, OG_r=OG_r, TSG_r=TSG_r, EG_r=EG_r)

        NEU_r = 1 - params.OG_r - params.TSG_r - params.EG_r
        for chrname, num_regions in self.regions.items():
            labels = np.random.choice([0, 1, 2, 3], size=num_regions, p=[params.OG_r, params.TSG_r, params.EG_r, NEU_r])
            for r,l in enumerate(labels):
                if l == 0:
                    # OG
                    self.delta[chrname].append(np.random.uniform(0, params.CN_coeff))
                    self.lam[chrname].append(params.SNV_coeff*((-1)**np.random.randint(0, 2)))
                    self.drivers[chrname].append(r)
                elif l == 1:
                    # TSG
                    self.delta[chrname].append(np.random.uniform(-params.CN_coeff, 0))
                    self.lam[chrname].append(params.SNV_coeff)
                    self.drivers[chrname].append(r)
                elif l == 2:
                    # EG
                    self.delta[chrname].append(0)
                    self.lam[chrname].append(params.SNV_coeff)
                    self.essential[chrname].append(r)
                else:
                    # NEU
                    self.lam[chrname].append(0)
                    self.delta[chrname].append(0)
        self.max_distinct_driv = int(self.max_distinct_driv_ratio * sum([len(self.drivers[chrname]) for chrname in self.drivers]))
        self.init_base_fit()
        self.init_max_fit()

class FittedHybridLibrary(BaseHybridLibrary):
    """Uses the same model as :code:`RandomHybridLibrary`, but the selection coefficients are given by the user.

    """
    def initialize(
        self, 
        filepath: Optional[str] = None,
        delta: Optional[Dict] = None,
        lam: Optional[Dict] = None,
    ):
        """Method used to initialize selection coefficients.

        Args:
            filepath (str, optional): Path to a tsv file containing the selection coefficients. The first entry is the chromosome name, followed by the selection coefficient pairs in order of reference regions. Each pair is of the form x,y where x is the CN coefficient and y is the SNV coefficient (:math:`\\delta` and :math:`\\lambda`).
            delta (dict, optional): Instead of passing a file, users can pass a dictionary where keys are chromosome names and values are an :math:`m_k`-length array of CN coefficients. Must also be passed with lam in the same form.
            lam (dict, optional): Instead of passing a file, users can pass a dictionary where keys are chromosome names and values are an :math:`m_k`-length array of SNV coefficients. Must also be passed with delta in the same form.
        
        """
        if delta is not None and lam is not None:
            if set(delta.keys()) != set(lam.keys()):
                raise ValueError("Provided delta and lam inputs should use the same chromosome names.")
            if len(set(delta.keys()) - set(self.chrom_lens.keys())) > 0:
                raise ValueError("Provided selection coefficient dictionary contains unrecognized chromosomes.")
            
            deltas, lam = {}, {}
            for chrname,v in delta.items():
                if not isinstance(v, (list, tuple, np.ndarray)) or type(v) != type(lam[chrname]):
                    raise ValueError("Provided selection coefficient dictionary values (both delta and lam) must be an array or list containing float values.")
                if len(v) != len(lam[chrname]):
                    raise ValueError(f"Number of deltas for {chrname} is different to the number of lambdas.")
                if len(v) > self.regions[chrname]:
                    raise ValueError(f"Number of deltas for {chrname} exceeds the number of regions specified by chrom_lens and region_len on library initialization.")
                for i in range(len(v)):
                    if not isinstance(v[i], (float, np.floating)) or not isinstance(lam[chrname][i], (float, np.floating)):
                        raise ValueError(f"Provided deltas are neither floats nor np.floats.")
                
                v1 = [float(x) for x in v]
                v2 = [float(x) for x in lam[chrname]]
                while len(v1) != self.regions[chrname]:
                    v1.append(0)
                    v2.append(0)
                deltas[chrname] = v1
                lam[chrname] = v2
        
        elif filepath is not None:
            assert os.path.isfile(filepath), 'File does not exist at specified path.'
            deltas, lam = {}, {}
            with open(filepath) as f:
                for line in f:
                    info = line.strip().split('\t')
                    chrname = info[0]
                    if chrname not in self.chrom_lens:
                        raise ValueError("Provided arm selection coefficients file contains unrecognized chromosome.")
                    deltas[chrname] = []
                    lam[chrname] = []
                    for x in info[1:]:
                        y = x.split(',')
                        deltas[chrname].append(float(y[0]))
                        lam[chrname].append(float(y[1]))

        else:
            raise ValueError("Must pass either filepath or delta and lam.")
        
        for chrname in self.regions:
            if chrname not in deltas:
                deltas[chrname] = [0 for _ in range(self.regions[chrname])]
                lam[chrname] = [0 for _ in range(self.regions[chrname])]
        self.delta = deltas
        self.lam = lam
        
        for chrname,v in deltas.items():
            for i,x in enumerate(v):
                if x != 0:
                    self.drivers[chrname].append(i)
                elif x == 0 and lam[chrname][i] != 0:
                    self.essential[chrname].append(i)

        self.max_distinct_driv = int(self.max_distinct_driv_ratio * sum([len(self.drivers[chrname]) for chrname in self.drivers]))
        self.init_base_fit()
        self.init_max_fit()


import numpy as np
import os

from sistem.data.utils import bin_region_values, get_mutated_basepairs, count_region_CNs
from sistem.lineage import CNA, SNV
from sistem.utilities.utilities import get_reg_id, get_bin_coordinates, sort_chrom_names
from sistem.utilities.IO import read_fasta

def save_mutations_from_tree(tree, out_dir, bin_size=None):
    log_mutations(tree, out_dir)
    save_CNPs(tree, out_dir, bin_size=bin_size)
    save_SNVs(tree, out_dir)

def save_CNPs(tree, out_dir, bin_size):
    """
    Saves the CNPs for the cells in a tree. Stores in two separate files for observed and ancestral cells. If region_len != bin_size, saves the region CNPs in a separate file.
    """
    chrom_lens = tree.root.library.chrom_lens
    region_len = tree.root.library.region_len
    chrnames = sort_chrom_names(chrom_lens.keys())

    if bin_size == None:
        bin_size = region_len

    region_coords = get_bin_coordinates(chrom_lens, region_len)

    observed_cells = [node for node in tree.iter_leaves()]
    ancestral_cells = [node for node in tree.iter_ancestors()]
    
    for group, keyword in zip([observed_cells, ancestral_cells], ['observed_', 'ancestral_']):
        region_profiles = {}
        for cell in group:
            region_profiles[cell.name] = count_region_CNs(cell)
        
        if bin_size == region_len:
            CNPs_to_tsv(region_profiles, region_coords, chrnames, os.path.join(out_dir, f'{keyword}CNPs.tsv'))
        else:
            #CNPs_to_tsv(region_profiles, region_coords, chrnames, os.path.join(out_dir, f'region_{keyword}CNPs.tsv'))

            bin_profiles = {}
            for cell in group:
                bin_profiles[cell.name] = bin_region_CNs(region_profiles[cell.name], region_len, bin_size)
            bin_coords = get_bin_coordinates(chrom_lens, bin_size)
            CNPs_to_tsv(bin_profiles, bin_coords, chrnames, os.path.join(out_dir, f'{keyword}CNPs.tsv'))

def CNPs_to_tsv(profiles, coords, chrnames, out_path):
    cell_names = list(profiles.keys())
    cell_names.sort()

    headers = ['Cell/Clone', 'Chrom', 'Start', 'End', 'Hap CN', 'Total CN']
    rows = []
    for chrname in chrnames:
        for i, (start, end) in enumerate(coords[chrname]):
            for cell in cell_names:
                CNa, CNb = round(profiles[cell][chrname][0][i], 3), round(profiles[cell][chrname][1][i], 3)
                CNtot = round(CNa + CNb, 3)
                rows.append(f'{cell}\t{chrname}\t{start}\t{end}\t{CNa},{CNb}\t{CNtot}\n')
    
    with open(out_path, 'w+') as f:
        f.write('\t'.join(headers) + '\n')
        for row in rows:
            f.write(row)

def bin_region_CNs(CNs, region_len, bin_size, total=False):
    """
    Partitions region CNs into continuous bins with the resulting bin CN being the average copy number of all regions belonging to that bin. The ratio bin_size/region_len may be a float value, in which case a weighted average is used and continuity of regions is maintained.

    Args:
        CNs (dict): Dictionary containing region copy numbers.
        region_len (int): Length of regions.
        bin_size (int): Length of bins.
    
    Returns:
        bin_CNs (dict): Binned copy numbers.
    """
    bin_CNs = {}
    r2b = bin_size/region_len

    for chrname in CNs:
        if total:
            bin_CNs[chrname] = bin_region_values(CNs[chrname], r2b)
        else:
            bin_CNs[chrname] = [bin_region_values(CNs[chrname][0], r2b), bin_region_values(CNs[chrname][1], r2b)]
    
    return bin_CNs


def save_SNVs(tree, out_dir, ref = None, alt_ref = None):
    """
    ref (str, None): Path to reference genome
    alt_ref (str, None):: Path to alt reference genome
    """
    chrom_lens = tree.root.library.chrom_lens
    region_len = tree.root.library.region_len
    chrnames = sort_chrom_names(chrom_lens.keys())

    observed_cells = [node for node in tree.iter_leaves()]
    mutated_bps = get_mutated_basepairs(observed_cells, chrnames, region_len)

    ref_handler, alt_ref_handler = None, None
    if ref:
        ref_handler = read_fasta(ref)
    if alt_ref:
        alt_ref_handler = read_fasta(alt_ref)

    profiles = create_SNV_profiles(observed_cells, mutated_bps, region_len, ref_handler=ref_handler, alt_ref_handler=alt_ref_handler)

    with open(os.path.join(out_dir, 'SNV_profiles.tsv'), 'w+') as f:
        f.write(f"Cell/Clone\tChrom\tPos\tGT\n")
        for i,(chrname, pos) in enumerate(mutated_bps):
            for cell in observed_cells:
                x = profiles[cell][i]
                f.write(f"{cell.name}\t{chrname}\t{pos}\t{x[0]}/{x[1]}\n")

def create_SNV_profiles(cells, mutated_bps, region_len, ref_handler = None, alt_ref_handler = None):
    n = len(mutated_bps)
    nucs = ['A', 'C', 'G', 'T']
    if ref_handler is None and alt_ref_handler is None:
        A_nucs = [str(np.random.choice(nucs)) for _ in range(n)]
        B_nucs = A_nucs
    else:
        A_nucs = []
        for chrname, pos in mutated_bps:
            bp = ref_handler[chrname][pos].seq.upper()
            if bp == 'N':
                A_nucs.append(str(np.random.choice(nucs)))
            else:
                A_nucs.append(bp)
        if alt_ref_handler is None:
            B_nucs = A_nucs
        else:
            B_nucs = []
            for i,(chrname, pos) in enumerate(mutated_bps):
                bp = alt_ref_handler[chrname][pos].seq.upper()
                if bp == 'N':
                    B_nucs.append(A_nucs[i])
                else:
                    B_nucs.append(bp)
    ref_nucs = [A_nucs, B_nucs]

    profiles = {cell: [['.', '.'] for _ in range(n)] for cell in cells}
    for cell in cells:
        for i,(chrname, pos) in enumerate(mutated_bps):
            r = pos // region_len
            p = pos % region_len
            for chromosome in cell.genome[chrname]:
                if profiles[cell][i][chromosome.allele] == '.':
                    for q in chromosome.seq:
                        if get_reg_id(q) == r:
                            profiles[cell][i][chromosome.allele] = ref_nucs[chromosome.allele][i]
                            break

                for q,positions in chromosome.SNVs.items():
                    if get_reg_id(q) == r:
                        if p in positions:
                            alt_nucs = nucs[:]
                            alt_nucs.remove(ref_nucs[chromosome.allele][i])
                            alt = alt_nucs[positions[p]]
                            profiles[cell][i][chromosome.allele] = alt
    return profiles

def log_mutations(tree, out_dir):
    with open(os.path.join(out_dir, 'CNA_events.tsv'), 'w+') as f:
        f.write(f'Cell/Clone\tType\tDriver\tChrom\tAllele\tArm\tStart Index\tLength\tCopies\tRef Region Indices\n')
        for node in tree.iter_descendants():
            for e in node.events:
                if isinstance(e, CNA):
                    if e.category == 'focal_amplification' or e.category == 'focal_deletion':
                        f.write(f"{node.name}\t{e.category}\t{e.driver}\t{e.chrom}\t{e.allele}\t{e.arm}\t{e.start}\t{e.end - e.start}\t{e.copies}\t{','.join([str(get_reg_id(g)) for g in e.gene_ids])}\n")
                    else:
                        f.write(f"{node.name}\t{e.category}\t{e.driver}\t{e.chrom}\t{e.allele}\t{e.arm}\tNone\tNone\tNone\tNone\n")
    
    with open(os.path.join(out_dir, 'SNV_events.tsv'), 'w+') as f:
        f.write(f'Cell/Clone\tDriver\tChrom\tAllele\tRegion Index\tRef Region Index\tPosition\tBP\n')
        for node in tree.iter_descendants():
            for e in node.events:
                if isinstance(e, SNV):
                    f.write(f"{node.name}\t{e.driver}\t{e.chrom}\t{e.allele}\t{e.index}\t{get_reg_id(e.region)}\t{e.position}\t{e.bp}\n")

def save_singlecell_readcounts(readcounts, out_path, chrom_lens, bin_size):
    chrnames = sort_chrom_names(chrom_lens.keys())
    bin_coords = get_bin_coordinates(chrom_lens, bin_size)
    with open(os.path.join(out_path, 'readcounts.tsv'), 'w+') as f:
        headers = ['Cell', 'Chrom', 'Start', 'End', 'Acount', 'Bcount']
        f.write('\t'.join(headers) + '\n')

        for chrname in chrnames:
            for i, (start, end) in enumerate(bin_coords[chrname]):
                for cell, rcs in readcounts.items():
                    f.write(f"{cell.name}\t{chrname}\t{start}\t{end}\t{rcs[chrname][0][i]}\t{rcs[chrname][1][i]}\n")

def save_clonal_readcounts(mutated_basepairs, readcounts, site_ids, out_path):
    with open(os.path.join(out_path, 'readcounts.tsv'), 'w+') as f:
        headers = ['Chrom', 'Pos'] + site_ids
        f.write('\t'.join(headers) + '\n')

        for i,(chrname,pos) in enumerate(mutated_basepairs):
            counts = [f"{readcounts[s][i][0]},{readcounts[s][i][1]}" for s in range(len(site_ids))]
            counts_combined = '\t'.join(counts)
            f.write(f"{chrname}\t{pos}\t{counts_combined}\n")

def site_CN_averages(tree, out_dir, bin_size):
    chrom_lens = tree.root.library.chrom_lens
    region_len = tree.root.library.region_len
    chrnames = sort_chrom_names(chrom_lens.keys())

    bin_coords = get_bin_coordinates(chrom_lens, bin_size)

    observed_clones = [node for node in tree.iter_leaves()]
    nsites = len(set([clone.site for clone in observed_clones]))
    
    for clone in observed_clones:
        if 'count' not in clone.info:
            raise ValueError("Leaves in Tree have not been initialized with a 'count' key/pair in the info attribute. Ensure you have generated the tree with GrowthSimulator.simulate_clonal_lineage, as this function does not work on trees generated with GrowthSimulator.simulate_singlecell_lineage.")

    cell_counts = {}
    for s in range(nsites):
        cell_counts[s] = {}
        for clone in observed_clones:
            if clone.site == s:
                cell_counts[s][clone] = clone.info['count']

    #Add diploid fractions to metastatic sites
    dip = tree.find('diploid')
    if dip is not None:
        dip_count = cell_counts[0][dip]
        prim_tumor_cells = sum(list(cell_counts[0].values())) - dip_count
        ratio = dip_count / prim_tumor_cells
        for s in range(1, nsites):
            met_tumor_cells = sum(list(cell_counts[s].values()))
            if met_tumor_cells > 0:
                num_dip = max(round(ratio*met_tumor_cells), 1)
                cell_counts[s][dip] = num_dip
    
    CN_averages = {s: {} for s in range(nsites)}
    for s in range(nsites):
        cur_clones = list(cell_counts[s].keys())
        weights = [cell_counts[s][clone] for clone in cur_clones]

        profiles = {}
        for clone in cur_clones:
            region_profiles = count_region_CNs(clone, total=True)
            profiles[clone] = bin_region_CNs(region_profiles, region_len, bin_size, total=True)
        
        for chrname in chrnames:
            CNs_flat = np.array([profiles[clone][chrname] for clone in cur_clones])
            chr_CN_avgs = np.average(CNs_flat, axis=0, weights=weights)
            CN_averages[s][chrname] = chr_CN_avgs
    
    site_ids = []
    for s in range(nsites):
        cur_clones = [clone.name for clone in cell_counts[s].keys()]
        if 'diploid' in cur_clones:
            cur_clones.remove('diploid')
        name = cur_clones[0]
        site_id = ''
        for char in name[:2]:
            if char.isupper():
                site_id += char
        site_ids.append(site_id)

    with open(os.path.join(out_dir, 'CN_averages.tsv'), 'w+') as f:
        f.write(f"Chrom\tStart\tEnd\tSite\tCN\n")
        for chrname in chrnames:
            for i, (start, end) in enumerate(bin_coords[chrname]):
                for s in range(nsites):
                    f.write(f"{chrname}\t{start}\t{end}\t{s}\t{CN_averages[s][chrname][i]}\n")


import numpy as np
import copy
import os
from collections import Counter
from typing import Optional, Union, Dict

from sistem.selection.base_library import BaseLibrary, Attributes
from sistem.anatomy.utils import assign_n_alt_drivers
from sistem.utilities.utilities import get_reg_id
from sistem.parameters  import Parameters, fill_params

class BaseArmLibrary(BaseLibrary):
    is_driver_region_model = False
    is_driver_SNV_model = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._arm_sizes = {}
        for chrname in self.regions.keys():
            if isinstance(self.arm_ratios, dict):
                arm_ratio = self.arm_ratios[chrname]
            else:
                arm_ratio = self.arm_ratios
            cent_idx = round(self.regions[chrname]*arm_ratio)
            # 'p' entry in _arm_sizes is also the cent index
            self._arm_sizes[chrname] = [cent_idx, self.regions[chrname] - cent_idx]
    
    def compute_fitness(self, clone):
        fitness = 1
        ploidy = clone.get_ploidy()
        for chrname in self.regions:
            for a in [0, 1]:
                s = self[chrname][a]
                n = clone.attributes.arm_counts[chrname][a] / self._arm_sizes[chrname][a]
                fitness *= (1 + s)**(n/ploidy)
        return fitness
    
    def init_base_fit(self):
        self.base_fit = 1
        for chrname in self.regions:
            for a in [0, 1]:
                s = self[chrname][a]
                self.base_fit *= (1 + s)

    def init_max_fit(self):
        def compute_fitness_from_counts(arm_counts, ploidy):
            if ploidy > self.max_ploidy or ploidy < self.min_ploidy:
                return -1
            fitness = 1
            for chrname in self.regions:
                for a in [0, 1]:
                    s = self[chrname][a]
                    n = arm_counts[chrname][a] / self._arm_sizes[chrname][a]
                    fitness *= (1 + s)**(n/ploidy)
            return fitness

        pos, neg = [], []
        for chrname,deltas in self.delta.items():
            for a,s in enumerate(deltas):
                if s > 0:
                    pos.append((chrname,a,s))
                elif s < 0:
                    neg.append((chrname,a,s))
        pos.sort(key=lambda x: x[2], reverse=True)
        neg.sort(key=lambda x: x[2])

        arm_counts = {chrname: [2*s for s in sizes] for chrname, sizes in self._arm_sizes.items()}
        cur_fit = self.base_fit
        start_nregions = self._ndiploid_regions
        cur_nregions = start_nregions*2
        count = 0
        while count < self.max_distinct_driv and (len(pos) > 0 or len(neg) > 0):
            fit1, fit2 = -1, -1
            if len(pos) > 0:
                chrname,a,s = pos[0]
                arm_counts[chrname][a] = self.max_region_CN*self._arm_sizes[chrname][a]
                nregions = cur_nregions + arm_counts[chrname][a] - (2*self._arm_sizes[chrname][a])
                fit1 = compute_fitness_from_counts(arm_counts, nregions/start_nregions)
                arm_counts[chrname][a] = 2*self._arm_sizes[chrname][a]
            if len(neg) > 0:
                chrname,a,s = neg[0]
                arm_counts[chrname][a] = 0
                nregions = cur_nregions - (2*self._arm_sizes[chrname][a])
                fit2 = compute_fitness_from_counts(arm_counts, nregions/start_nregions)
                arm_counts[chrname][a] = 2*self._arm_sizes[chrname][a]
            if fit1 == -1 and fit2 == -1:
                break
            elif fit1 > fit2:
                chrname,a,s = pos.pop(0)
                arm_counts[chrname][a] = self.max_region_CN*self._arm_sizes[chrname][a]
                cur_nregions = cur_nregions + arm_counts[chrname][a] - (2*self._arm_sizes[chrname][a])
                cur_fit = fit1
            else:
                chrname,a,s = neg.pop(0)
                arm_counts[chrname][a] = 0
                cur_nregions = cur_nregions - (2*self._arm_sizes[chrname][a])
                cur_fit = fit2
            count += 1
        self.max_fit = cur_fit
    
    def check_viability(self, clone):
        """
        Checks if clone passes viability checkpoints based on mutated driver stats. 

        Args:
            clone (Clone): The clone object in question.

        Returns:
            (bool): True if passes, False if not.
        
        """
        max_chrom_CNs = [0]
        for chrname in self.regions:
            c = Counter()
            for chromosome in clone.genome[chrname]:
                c.update(Counter(chromosome.seq))
            if len(c):
                max_chrom_CNs.append(c.most_common()[0][1])
        max_CN = max(max_chrom_CNs)
        
        p = 2**(clone.genome.nWGD + 1)
        nmutated_arms = 0
        for chrname in self.regions:
            for a in [0, 1]:
                r = clone.attributes.arm_counts[chrname][a]/(self._arm_sizes[chrname][a]*p)
                if r > 1.5 or r < 0.5:
                    nmutated_arms += 1

        ploidy = clone.get_ploidy()

        if nmutated_arms > self.max_distinct_driv:
            return False
        elif ploidy > self.max_ploidy:
            return False
        elif ploidy < self.min_ploidy:
            return False
        elif max_CN > self.max_region_CN:
            return False
        else:
            return True

    def update_stats(self, clone, chromosome, start, end, mag=1):
        for r in chromosome.seq[start:end]:
            if r < self._arm_sizes[chromosome.name][0]:
                clone.attributes.arm_counts[chromosome.name][0] += mag
            else:
                clone.attributes.arm_counts[chromosome.name][1] += mag
    
    def init_attributes(self, clone):
        arm_counts = {chrname: [0, 0] for chrname in self.regions}
        if clone.genome is not None:
            for chromosome in clone.genome.get_chromosomes():
                chrname = chromosome.name
                for r in chromosome.seq:
                    if r < self._arm_sizes[chrname][0]:
                        arm_counts[chrname][0] += 1
                    else:
                        arm_counts[chrname][1] += 1
        clone.attributes = Attributes(arm_counts=arm_counts)

    def get_driver_start_regions(self, cell, chromosome, size):
        n = len(chromosome) - size
        return list(range(n))

    def get_passenger_start_regions(self, cell, chromosome, size):
        sequence = [get_reg_id(r) for r in chromosome.seq]
        n = len(sequence) - size + 1
        blregions = None
        if chromosome.name in cell.attributes.blacklisted_regions:
            if chromosome.homolog_id in cell.attributes.blacklisted_regions[chromosome.name]:
                blregions = set(cell.attributes.blacklisted_regions[chromosome.name][chromosome.homolog_id])
        if blregions is None:
            return list(range(n))

        starts = []
        i = 0

        while i < n:
            hits = [j for j,r in enumerate(sequence[i:i+size], i) if r in blregions]
            if len(hits) > 0:
                i = max(hits) + 1
            else:
                while i < n and sequence[i+size-1] not in blregions:
                    starts.append(i)
                    i += 1
                    
        return starts
    
    def create_metastatic_libraries(self, nsites, alter_prop, CN_coeff, method='random', dists=None):
        if method == 'random':
            return self.create_metastatic_libraries_random(nsites, alter_prop, CN_coeff)
        elif method == 'distance':
            if dists is None:
                raise ValueError("Cannot create metastatic libraries with the 'distance' method when dists has not been initialized.")
            return self.create_metastatic_libraries_distance(nsites, alter_prop, CN_coeff, dists)
        else:
            raise ValueError("Method not recognized.")
    
    def create_metastatic_libraries_random(self, nsites, alter_prop, CN_coeff):
        init_drivers = [(chrname, a) for chrname in self.regions for a in [0, 1]]

        ndrivers = len(init_drivers)
        libraries = [self]
        for _ in range(1, nsites):
            new_lib = copy.deepcopy(self)
            n_alt_drivers = min(np.random.binomial(ndrivers, alter_prop), ndrivers)
            idxs = np.random.choice(range(ndrivers), n_alt_drivers, replace=False)
            alt_drivers = [init_drivers[i] for i in idxs]
            for chrname, a in alt_drivers:
                new_lib.delta[chrname][a] = np.random.uniform(-CN_coeff, CN_coeff)
            new_lib.init_base_fit()
            new_lib.init_max_fit()
            libraries.append(new_lib)

        return libraries

    def create_metastatic_libraries_distance(self, nsites, alter_prop, CN_coeff, dists):
        init_drivers = [(chrname, a) for chrname in self.regions for a in [0, 1]]

        ndrivers = len(init_drivers)
        n_alt_drivers = np.random.binomial(ndrivers, alter_prop)
        dists = np.rint(dists*n_alt_drivers).astype(int)
        idxs = np.triu_indices(nsites)
        dmatrix = np.zeros((nsites, nsites), dtype=int)
        dmatrix[idxs] = dists[:, 0]
        dmatrix = dmatrix + dmatrix.T
        A, score = assign_n_alt_drivers(dmatrix, ndrivers)

        libraries = [self]
        unique = list(set.union(*[x for x in A.values()]))
        idxs = np.random.choice(range(ndrivers), len(unique), replace=False)
        alt_drivers = [init_drivers[i] for i in idxs]
        new_drivers = {}
        for i,(chrname, a) in zip(unique, alt_drivers):
            new_delta = np.random.uniform(-CN_coeff, CN_coeff)
            new_drivers[i] = (chrname, a, new_delta)

        for s in range(1, nsites):
            new_lib = copy.deepcopy(self)
            for i in A[s]:
                chrname, a, delta = new_drivers[i]
                new_lib.delta[chrname][a] = delta
            new_lib.init_base_fit()
            new_lib.init_max_fit()
            libraries.append(new_lib)

        return libraries

class RandomArmLibrary(BaseArmLibrary):
    """The chromosome-arm selection model with randomly generated selection coefficients.
    
    Under this model, each chromosome-arm is assigned a random selection coefficient in the range (-CN_coeff, CN_coeff). Arms with a positive coefficient correspond to those with a greater balance of OGs to TSGs, while arms with a negative coefficient will have a greater balance of TSGs to OGs. The fitness :math:`s_a(c)` of cell :math:`c` in site :math:`a` is computed as

    .. math::
    
        s_a(c) = \\prod_{k=1}^K \\prod_{arm \\in \\{p,q\\}} (1 + \\delta_{k,arm})^{x_{k,arm} / p_c},

    where :math:`K` is the number of chromosomes, :math:`\\delta` are selection coefficients, :math:`x_{k,arm}` is the average copy number of chromosome arm :math:`k,arm` in cell :math:`c`, and :math:`p_c` is cell ploidy.

    """
    def initialize(
        self, 
        params: Optional[Parameters] = None, 
        CN_coeff: Optional[float] = None
    ):
        """Method used to initialize selection coefficients. See :ref:`Parameters <parameters>` for an explanation of the parameters.

        Args:
            params (Parameters, optional):
            CN_coeff (float, optional):
        
        """
        params = fill_params(params, CN_coeff=CN_coeff)

        arm_deltas = {}
        for chrname in self.regions:
            arm_deltas[chrname] = list(np.random.uniform(-params.CN_coeff, params.CN_coeff, 2))
        self.delta = arm_deltas
        self.max_distinct_driv = int(self.max_distinct_driv_ratio * (2*len(self.regions)))
        self.init_base_fit()
        self.init_max_fit()

class FittedArmLibrary(BaseArmLibrary):
    """Uses the same model as :code:`RandomArmLibrary`, but the selection coefficients are given by the user.
    
    """
    def initialize(
        self, 
        filepath: Optional[str] = None,
        delta: Optional[Dict] = None,
    ):
        """Method used to initialize selection coefficients.

        Args:
            filepath (str, optional): Path to a tsv file containing the selection coefficients. The first entry is the chromosome name, followed by the selection coefficient of the short and long arm, respectively.
            delta (dict, optional): Instead of passing a file, users can pass a dictionary where keys are chromosome names and values are a 2-length array of coefficients.
        
        """
        if delta is not None:
            if len(set(delta.keys()) - set(self.chrom_lens.keys())) > 0:
                raise ValueError("Provided selection coefficient dictionary contains unrecognized chromosomes.")
            for i,v in delta.items():
                if not isinstance(v, (list, tuple, np.ndarray)):
                    raise ValueError("Provided selection coefficient dictionary values must be an array, list, or tuple containing 2 float values.")
                v = list(v)
                if len(v) != 2 or not (isinstance(v[0], (float, np.floating)) or isinstance(v[1], (float, np.floating))):
                    raise ValueError("Provided selection coefficient dictionary values must be an array, list, or tuple containing 2 float values.")
            arm_deltas = {i: list(v) for i,v in delta.items()}
        elif filepath is not None:
            assert os.path.isfile(filepath), 'File does not exist at specified path'
            arm_deltas = {}
            with open(filepath) as f:
                for line in f:
                    info = line.strip().split('\t')
                    chrname = info[0]
                    if chrname not in self.chrom_lens:
                        raise ValueError("Provided arm selection coefficients file contains unrecognized chromosome.")
                    arm_deltas[chrname] = [float(info[1]), float(info[2])]
        else:
            raise ValueError("Must pass either filepath or delta.")
        
        for chrname in self.chrom_lens:
            if chrname not in arm_deltas:
                arm_deltas[chrname] = [0, 0]
        self.delta = arm_deltas
        self.max_distinct_driv = int(self.max_distinct_driv_ratio * (2*len(self.regions)))
        self.init_base_fit()
        self.init_max_fit()



import os
import subprocess
import numpy as np
import multiprocessing
from Bio import SeqIO
from scipy.stats import poisson
from typing import Optional

from sistem.utilities.utilities import iter_by_chunk, get_reg_id, sort_chrom_names, check_dependencies
from sistem.data.utils import gen_coverage, get_alpha_beta, bin_region_readcounts, get_mutated_basepairs, count_region_CNs
from sistem.data.sequence import build_cell_ref
from sistem.data.profiles import save_singlecell_readcounts, save_clonal_readcounts
from sistem.lineage import Tree
from sistem.parameters import Parameters, fill_params
from sistem.utilities.IO import read_fasta
from sistem.utilities.utilities import init_mbit

# Draw readcounts for each bin
def draw_readcounts_nonuniform(num_windows, Aa, Bb, mean_rc, interval=5):
    cov_scales = gen_coverage(num_windows, interval, Aa, Bb)
    readcounts = [round(poisson.rvs(mean_rc*x)) for x in cov_scales]
    return readcounts

def gen_reads_cell(
    cell, 
    readcounts, 
    out_dir, 
    ref_path, 
    alt_ref_path, 
    read_len, 
    seq_error
):
    ref = read_fasta(ref_path)
    if alt_ref_path is None:
        refs = [ref, ref]
    else:
        alt_ref = read_fasta(alt_ref_path)
        refs = [ref, alt_ref]

    chrom_lens = cell.library.chrom_lens
    regions = cell.library.regions
    region_len = cell.library.region_len

    max_region_id = max(regions.values())
    init_mbit(max_region_id.bit_length())

    last_region_lens = {chrname: int(l % region_len) for chrname,l in chrom_lens.items()}

    read_dir = os.path.join(out_dir, 'reads')
    prefix = os.path.join(read_dir, f"{cell.name}")
    prefixes = [f"{prefix}_allele0", f"{prefix}_allele1"]
    for cur_prefix in prefixes:
        open(f"{cur_prefix}.read1.fastq.gz", 'w+').close()
        open(f"{cur_prefix}.read2.fastq.gz", 'w+').close()

    for chrname in chrom_lens:
        for chromosome in cell.genome[chrname]:
            if len(chromosome) > 0:
                cur_ref = refs[chromosome.allele]
                cur_prefix = prefixes[chromosome.allele]
                cur_readcounts = readcounts[chromosome]

                chrom_fa_path = f"{cur_prefix}.{chrname}_{chromosome.homolog_id}.fa"
                build_cell_ref(chromosome, cur_ref, regions[chrname], region_len, chrom_fa_path)

                start,end = 0,0
                for i,r in enumerate(chromosome.seq):
                    q = get_reg_id(r)
                    rc = cur_readcounts[i]
                    if q == regions[chrname] - 1:
                        end += last_region_lens[chrname]
                    else:
                        end += region_len
                    
                    region_fa_path = f"{chrom_fa_path[:-3]}_{int(start)}_{int(end)}.fa"
                    with open(region_fa_path, 'w+') as f:
                        call = subprocess.run(['samtools', 'faidx', chrom_fa_path, f"{chrname}:{int(start)}-{int(end)}"], stdout=f)

                    for record in SeqIO.parse(region_fa_path, 'fasta'):
                        total_N = record.seq.count('N')
                    total_bp = end - start + 1
                    N_ratio = total_N / total_bp
                    rc = int(rc*(1-N_ratio))

                    if rc > 0:
                        call = subprocess.run(['dwgsim', '-H', '-o', '1', '-N', str(rc), '-1', str(read_len), '-2', str(read_len), '-e', str(seq_error), '-E', str(seq_error), region_fa_path, region_fa_path[:-3]], capture_output=True, text=True)

                        with open(f"{cur_prefix}.read1.fastq.gz", 'a') as f1, open(f"{cur_prefix}.read2.fastq.gz", 'a') as f2:
                            call = subprocess.run(['cat', region_fa_path[:-3] + '.bwa.read1.fastq.gz'], stdout=f1)
                            call = subprocess.run(['cat', region_fa_path[:-3] + '.bwa.read2.fastq.gz'], stdout=f2)
                        os.remove(region_fa_path[:-3] + '.bwa.read1.fastq.gz')
                        os.remove(region_fa_path[:-3] + '.bwa.read2.fastq.gz')
                        os.remove(region_fa_path[:-3] + '.mutations.txt')
                        os.remove(region_fa_path[:-3] + '.mutations.vcf')
                    os.remove(region_fa_path)
                    start = end
                os.remove(chrom_fa_path)
                os.remove(f"{chrom_fa_path}.fai")
    
    with open(prefix + '.read1.fastq.gz', 'w+') as f1, open(prefix + '.read2.fastq.gz', 'w+') as f2:
        call = subprocess.run(['cat', prefix + '_allele0.read1.fastq.gz', prefix + '_allele1.read1.fastq.gz'], stdout=f1)
        call = subprocess.run(['cat', prefix + '_allele0.read2.fastq.gz', prefix + '_allele1.read2.fastq.gz'], stdout=f2)
    os.remove(prefix + '_allele0.read1.fastq.gz')
    os.remove(prefix + '_allele1.read1.fastq.gz')
    os.remove(prefix + '_allele0.read2.fastq.gz')
    os.remove(prefix + '_allele1.read2.fastq.gz')

def gen_reads(
    tree: Tree,
    params: Optional[Parameters] = None,
    out_dir: Optional[str] = None,
    ref: Optional[str] = None,
    alt_ref: Optional[str] = None,
    coverage: Optional[float] = None,
    bin_size: Optional[int] = None,
    read_len: Optional[int] = None,
    lorenz_y: Optional[float] = None,
    num_processors: Optional[int] = None,
    lorenz_x: float = 0.5
):
    """Given a Tree object returned by GrowthSimulator.simulate_singlecell_lineage, generates synthetic paired-end DNA sequencing reads for each cell. Requires the *dwgsim* and *samtools* binaries be available in the user's :code:`$PATH` variable. Will create two fastq files for each cell located in the 'reads' directory within the provided output directory. See :ref:`Parameters <parameters>` for an explanation of the parameters.

    Args:
        tree (Tree):
        params (Parameters, optional): 
        out_dir (str, optional): 
        ref (str, optional): 
        alt_ref (str, optional): 
        coverage (float, optional): 
        bin_size (int, optional): 
        read_len (int, optional): 
        lorenz_y (float, optional): 
        num_processors (int, optional): 
    """
    params = fill_params(params, out_dir=out_dir, ref=ref, alt_ref=alt_ref, coverage=coverage, bin_size=bin_size, read_len=read_len, lorenz_y=lorenz_y, num_processors=num_processors)

    if params.ref is None:
        raise ValueError(f"Must pass a reference genome.")

    read_dir = os.path.join(params.out_dir, 'reads')
    if not os.path.isdir(read_dir):
        os.makedirs(read_dir)

    check_dependencies(['samtools', 'dwgsim'])

    readcounts = gen_readcounts_singlecell(tree, params=params, lorenz_x=lorenz_x)
    leaves = list(tree.iter_leaves())

    if params.num_processors == 1:
        for cell in leaves:
            gen_reads_cell(cell, readcounts[cell], params.out_dir, params.ref, params.alt_ref, params.read_len, params.seq_error)
    else:
        with multiprocessing.Pool(processes=params.num_processors) as pool:
            args = [(cell, readcounts[cell], params.out_dir, params.ref, params.alt_ref, params.read_len, params.seq_error) for cell in leaves]
            pool.starmap(gen_reads_cell, args)
            
def gen_readcounts_singlecell(
    tree: Tree,
    params: Optional[Parameters] = None,
    out_dir: Optional[str] = None,
    coverage: Optional[float] = None,
    bin_size: Optional[int] = None,
    read_len: Optional[int] = None,
    lorenz_y: Optional[float] = None,
    lorenz_x: float = 0.5
):
    """Given a Tree object returned by GrowthSimulator.simulate_singlecell_lineage, compiles cell-specific region-level copy numbers and draws allele-specific read counts. These will automatically be saved in the output directory. See :ref:`Parameters <parameters>` for an explanation of the parameters.

    Args:
        tree (Tree):
        params (Parameters, optional): 
        out_dir (str, optional): 
        coverage (float, optional): 
        bin_size (int, optional): 
        read_len (int, optional): 
        lorenz_y (float, optional): 

    Returns:
        dict: A dictionary where keys are cells and values are a dictionary containing the region-level allele-specific copy numbers.
    """
    params = fill_params(params, out_dir=out_dir, coverage=coverage, bin_size=bin_size, read_len=read_len, lorenz_y=lorenz_y)
    region_len = tree.root.library.region_len
    regions = tree.root.library.regions
    chrom_lens = tree.root.library.chrom_lens
    mean_rc = (region_len*(params.coverage/2)) / (2*params.read_len)
    regions_per_bin = params.bin_size / region_len
    [Aa, Bb] = get_alpha_beta(lorenz_x, params.lorenz_y)

    last_region_lens = {chrname: int(l % region_len) for chrname,l in chrom_lens.items()}

    leaves = list(tree.iter_leaves())
    gt_readcounts = {}
    region_readcounts = {cell: {} for cell in leaves}

    for cell in leaves:
        full_readcounts = {chrname: [[0 for i in range(regions[chrname])], [0 for i in range(regions[chrname])]] for chrname in chrom_lens}
        for chrname in chrom_lens:
            for chromosome in cell.genome[chrname]:
                if params.lorenz_y == 0.5: # Uniform Coverage
                    cur_readcounts = [round(poisson.rvs(mean_rc)) for i in range(len(chromosome.seq))]

                else: # Non-uniform Coverage
                    cur_readcounts = draw_readcounts_nonuniform(len(len(chromosome.seq)), Aa, Bb, mean_rc, interval=5)
                region_readcounts[cell][chromosome] = cur_readcounts

                for i,r in enumerate(chromosome.seq):
                    q = get_reg_id(r)
                    rc = cur_readcounts[i]
                    if q == regions[chrname] - 1:
                        rc = rc * (last_region_lens[chrname] / region_len)
                    full_readcounts[chrname][chromosome.allele][q] += rc


        binned_readcounts = {chrname: [] for chrname in chrom_lens}
        for chrname in chrom_lens:
            for allele in [0, 1]:
                binned_readcounts[chrname].append(bin_region_readcounts(full_readcounts[chrname][allele], regions_per_bin))
        gt_readcounts[cell] = binned_readcounts
    save_singlecell_readcounts(gt_readcounts, params.out_dir, tree.root.library.chrom_lens, params.bin_size)

    return region_readcounts

def gen_readcounts_bulk(
    tree: Tree,
    params: Optional[Parameters] = None,
    out_dir: Optional[str] = None,
    coverage: Optional[float] = None
):
    """Given a Tree object returned by GrowthSimulator.simulate_clonal_lineage, draws a total and variant read count for each site at every SNV position in the observed cells. These will automatically be saved in the output directory. See :ref:`Parameters <parameters>` for an explanation of the parameters.


    Args:
        tree (Tree):
        params (Parameters, optional): 
        out_dir (str, optional): 
        coverage (float, optional): 
    """
    params = fill_params(params, out_dir=out_dir, coverage=coverage)
    chrom_lens = tree.root.library.chrom_lens
    region_len = tree.root.library.region_len
    chrnames = sort_chrom_names(chrom_lens.keys())

    observed_clones = [node for node in tree.iter_leaves()]
    mutated_basepairs = get_mutated_basepairs(observed_clones, chrnames, region_len)

    nsites = len(set([clone.site for clone in observed_clones]))

    for clone in observed_clones:
        if 'count' not in clone.info:
            raise ValueError("Leaves in Tree have not been initialized with a 'count' key/pair in the info attribute. Ensure you have run GrowthSimulator.simulate_clonal_lineage.")
    
    #Count the number of cells sampled from each clone, separated by site
    cell_counts = {}
    for s in range(nsites):
        cell_counts[s] = {}
        for clone in observed_clones:
            if clone.site == s:
                cell_counts[s][clone] = clone.info['count']

    #Add diploid fractions to metastatic sites
    dip = tree.find('diploid')
    if dip is not None:
        dip_count = cell_counts[0][dip]
        prim_tumor_cells = sum(list(cell_counts[0].values())) - dip_count
        ratio = dip_count / prim_tumor_cells
        for s in range(1, nsites):
            met_tumor_cells = sum(list(cell_counts[s].values()))
            if met_tumor_cells > 0:
                num_dip = max(round(ratio*met_tumor_cells), 1)
                cell_counts[s][dip] = num_dip
    
    #Compute readcounts
    readcounts = {s: [] for s in range(nsites)}
    for s in range(nsites):
        ncells = sum(list(cell_counts[s].values()))
        cur_clones = list(cell_counts[s].keys())
        if ncells == 0:
            continue
        props = [cell_counts[s][clone] / ncells for clone in cur_clones]
        for chrname,pos in mutated_basepairs:
            r = pos // region_len
            p = pos % region_len

            fraction = 0
            CNs = []
            for clone,prop in zip(cur_clones, props):
                x_ref, x_alt = 0, 0
                for chromosome in clone.genome[chrname]:
                    for q in chromosome.seq:
                        if get_reg_id(q) == r:
                            x_ref += 1
                            if q in chromosome.SNVs:
                                if p in chromosome.SNVs[q]:
                                    x_ref -= 1
                                    x_alt += 1
                fraction += (prop * (x_alt / (x_ref + x_alt + 1e-8)))
                CNs.append(x_ref + x_alt)

            scale = np.mean(CNs) / 2
            exp_rc = params.coverage * scale
            tot_rc = poisson.rvs(exp_rc)
            var_rc = round(tot_rc*fraction)
            readcounts[s].append([tot_rc, var_rc])
    
    site_ids = []
    for s in range(nsites):
        cur_clones = [clone.name for clone in cell_counts[s].keys()]
        if 'diploid' in cur_clones:
            cur_clones.remove('diploid')
        name = cur_clones[0]
        site_id = ''
        for char in name[:2]:
            if char.isupper():
                site_id += char
        site_ids.append(site_id)

    save_clonal_readcounts(mutated_basepairs, readcounts, site_ids, params.out_dir)





import numpy as np
from sklearn.preprocessing import MinMaxScaler

from collections import defaultdict
import os

def get_num_migrations(clone, p):
    nmigrations = np.random.binomial(clone.popsize, p)
    return nmigrations

def draw_migrations(nmigrations, tprobs):
    migration_counts = np.random.multinomial(nmigrations, tprobs / sum(tprobs))
    migrations = (migration_counts > 0).astype(int)
    return migrations

def create_random_distances(nsites):
    points = [np.random.uniform(size=2) for i in range(nsites)]
    dists = []

    for i in range(nsites):
        for j in range(i, nsites):
            dists.append(np.linalg.norm(points[i] - points[j]))
    dists = np.array(dists).reshape(-1, 1)
    norm_dists = MinMaxScaler().fit_transform(dists)

    return points, norm_dists

def compute_diff(A, B):
    '''
    Returns the number of elements unique to A or B. 
    '''
    return len(A.union(B)) - len(A.intersection(B))

def compute_score(A, D):
    score = 0
    for i in range(D.shape[0] - 1):
        for j in range(i, D.shape[0]):
            diff = compute_diff(A[i], A[j])
            score += np.abs(D[i,j] - diff)
    return score

def get_average_dist(A, D, i):
    dists = []
    for j in range(D.shape[0]):
        if i != j:
            diff = compute_diff(A[i], A[j])
            dists.append(np.abs(D[i,j] - diff))
    return np.mean(dists)

def get_total_dist(A, D, i):
    dists = []
    for j in range(D.shape[0]):
        if i != j:
            diff = compute_diff(A[i], A[j])
            dists.append(np.abs(D[i,j] - diff))
    return np.sum(dists)

def find_improvement(A, D, used, unused, score):
    order = [i for i in range(1, D.shape[0])]
    order.sort(key = lambda x: get_average_dist(A, D, x), reverse=True)
    candidates = used[:]
    #print(order)
    if len(unused):
        candidates.append(unused[0])
    for i in order:
        scores, els = [], []
        for c in candidates:
            if c not in A[i]:
                A[i].add(c)
                els.append(c)
                scores.append(compute_score(A, D))
                A[i].remove(c)
                #print(i, c, scores[-1])
        #print(i, scores, els)
        if len(scores):
            idx = np.argmin(scores)
            best = scores[idx]
            if best < score:
                A[i].add(els[idx])
                if len(unused) and els[idx] == unused[0]:
                    unused.pop(0)
                    used.append(els[idx])
                return best
    return score

def assign_n_alt_drivers(D, num_elements):
    nsites = D.shape[0]
    A = {i: set() for i in range(nsites)}
    scores = [compute_score(A, D)]
    used = []
    unused = [e for e in range(num_elements)]
    while len(scores) <= 1 or scores[-1] < scores[-2]:
        next_score = find_improvement(A, D, used, unused, scores[-1])
        #print(next_score, used, unused, A)
        scores.append(next_score)
    return A, scores[-1]

def get_dist_matrix(A):
    if isinstance(A, list):
        alt = {i: set(v) for i,v in enumerate(A)}
    else:
        alt = A
    mat = np.zeros((len(A), len(A)), dtype=int)
    for i in range(len(A)):
        for j in range(i, len(A)):
            if i == j:
                mat[i,i] = 0
            else:
                diff = compute_diff(alt[i], alt[j])
                mat[i,j] = mat[j,i] = diff
    return mat

def generate_migration_graph(tree, site_ids):
    edges = defaultdict(int)
    for node in tree.traverse():
        psite = site_ids[node.site]
        for child in node.children:
            csite = site_ids[child.site]
            if psite != csite:
                edges[f'{psite}-{csite}'] += 1
    return dict(edges)

def save_migration_graph(out_dir, edges):
    with open(os.path.join(out_dir, f'migration_graph.tsv'), 'w+') as f:
        for i,v in edges.items():
            f.write(f'{i}\t{v}\n')


import numpy as np
from typing import Optional, Union, List

from sistem.anatomy.base_anatomy import BaseAnatomy
from sistem.selection import BaseLibrary

class GenotypeAnatomy(BaseAnatomy):
    """The genotype migration model.
    
    In this model, the probability of cell :math:`c` migrating from site :math:`a` to site :math:`b` increases as :math:`c` gains beneficial mutations with respect to the selection landscape of :math:`b`. In particular, distance is defined as

    .. math::

        d(a,b) = \\Big(\\log{\\frac{s_b(c)}{\\hat{s}_{b}}}\\Big)^{-1},

    where :math:`\\hat{s}_{b}` is the fitness of a non-mutated cell in site :math:`b`. In other words, the probability that cell :math:`c` successfully migrates to :math:`a` site :math:`b` increases as the cell acquires beneficial mutations. When used in conjunction with site-specific selection libraries, migration probabilities can also reflect compatibility with the the fitness landscape :math:`b`.

    """
    def __init__(self, libraries: Optional[Union[BaseLibrary, List[BaseLibrary]]], **kargs):
        super().__init__(libraries=libraries, **kargs)
        self.initialized = True

    def get_oneway_transition_probabilities(self, clone):
        s = clone.site
        probs = []
        for i in range(len(self.libraries)):
            if i == s:
                probs.append(0)
            else:
                multiplier = max(np.log(self.libraries[i].compute_fitness(clone) / self.libraries[i].base_fit), 1)
                probs.append(multiplier * self.eps)
        return np.array(probs)
    
    def get_total_transition_probability(self, clone):
        sprobs = self.get_oneway_transition_probabilities(clone)
        tot_prob = 1 - np.prod([1 - x for x in sprobs])
        return tot_prob


import numpy as np
import sys
import itertools
import shutil
from typing import Optional, List

class DependencyError(Exception):
    """Exception raised when required dependencies are missing."""
    pass

def check_dependencies(commands: Optional[List[str]] = None):
    missing = []
    for cmd in commands:
        if shutil.which(cmd) is None:
            missing.append(cmd)
    if missing:
        missing_combined = ', '.join(missing)
        raise DependencyError(f"The following binaries must be installed and callable: {missing_combined}.")
    return True

def compute_growth_rate(N0, Nt, tgen):
    '''
    Computes the exponential growth rate.

    Args:
        N0 (int): Population size at time t0.
        Nt (int): Population size at time t.
        tgen (int): Number of generations between t0 and t.
    
    Returns:
        (float): Exponential growth rate.
    '''
    return (np.log(Nt) - np.log(N0)) / tgen
    
def sort_chrom_names(chrnames):
    all_contains_prefix = np.all(np.array([chrname[:3] for chrname in chrnames]) == 'chr')
    all_missing_prefix = np.all(np.array([chrname[:3] for chrname in chrnames]) != 'chr')
    # If some have prefix 'chr' and others do not, naming is inconsistent and return what was inputted.
    if not (all_contains_prefix or all_missing_prefix):
        return chrnames
    
    if all_contains_prefix:
        # If any chrnames are are just 'chr', then return.
        if not np.all(np.array([len(chrname) for chrname in chrnames]) > 3):
            return chrnames
        chrnames = [chrname[3:] for chrname in chrnames]
    
    int_chrnames = [int(x) for x in chrnames if x.isdigit()]
    XY_chrnames = [x for x in chrnames if x[0] == 'X' or x[0] == 'Y']
    alt_chrnames = [x for x in chrnames if not (x.isdigit() or x[0] == 'X' or x[0] == 'Y')]
    int_chrnames.sort()
    XY_chrnames.sort()
    alt_chrnames.sort()
    sorted_chrnames = int_chrnames + XY_chrnames + alt_chrnames
    if all_contains_prefix:
        sorted_chrnames = [f'chr{x}' for x in sorted_chrnames]
    return sorted_chrnames
    
def get_bin_coordinates(chrom_lens, bin_size):
    bin_coords = {chrname: [] for chrname in chrom_lens}
    for chrname, chrlen in chrom_lens.items():
        nbins = int(chrlen // bin_size)
        rem = chrlen % bin_size
        if rem > 0:
            nbins += 1
        for i in range(nbins):
            start = int(i*bin_size + 1)
            end = int(min((i+1)*bin_size, chrlen))
            bin_coords[chrname].append((start, end))
    return bin_coords
    
def get_size(obj, seen=None):
    """Recursively finds size of objects"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size


def combine_dicts(*dicts):
    """
    Given a set of 2-depth dictionaries of sets, combines all keys and subkeys into a single dictionary with unique values present in any of the inputs.

    Args:
        *dicts (iterable): Collection of dictionaries.
    
    Returns:
        combined (dict): Output dictionary.
    """

    combined = {}
    for d in dicts:
        for key in d:
            if key not in combined:
                combined[key] = {}
            for subkey in d[key]:
                if subkey not in combined[key]:
                    combined[key][subkey] = set()
                combined[key][subkey].update(d[key][subkey])

    #for key, subkeys in combined.items():
    #    for subkey in subkeys:
    #        combined[key][subkey] = list(set(combined[key][subkey]))
    
    return combined

def iter_by_chunk(iterable, chunksize):
    return itertools.zip_longest(*[iter(iterable)] * chunksize)

# Initialize the MBIT to be one significant bit higher than what is needed to represent the largest region id. All bits to the left of MBIT will be used as a unique id.
def init_mbit(val):
    global MBIT
    val = val << 1
    MBIT = val

# Given any integer x, return the unique id.
def get_reg_count(x):
    mask = ~((1 << (MBIT + 1)) - 1)
    y = (x & mask) >> (MBIT + 1)
    return y

# Given any integer x, return the true region id
def get_reg_id(x):
    mask = ((1 << (MBIT + 1)) - 1)
    r = (x & mask)
    return r

# Given true region id x and a unique identifier b, return x with b prepended to it before the MBIT.
def update(x, b):
    mask = ((1 << (MBIT + 1)) - 1)
    x_prime = (b << (MBIT + 1)) | (x & mask)
    return x_prime

# Given integer x, return x with MBIT flipped
#def inc(x):
#    return update(x, 1)


import numpy as np

def extract_largest_clones(gs):
    '''
    Finds the largest clone in each anatomical site based on popsize and returns it
    '''
    largest = []
    for s,clones in gs.clones.items():
        if len(clones):
            c = max(clones, key = lambda x: x.popsize)
            largest.append(c)
        else:
            largest.append(None)
    return largest

def arm_CNratios_single(cell):
    CNratios = {}
    for chrname in cell.genome.chrom_names:
        CNratios[chrname] = []
        for a in [0, 1]:
            CNrat = cell.attributes.arm_counts[chrname][a]/(cell.library.arm_sizes[chrname][a])
            CNratios[chrname].append(CNrat)
    return CNratios

def arm_CNratios_bulk(clones):
    popsizes = [clone.popsize for clone in clones]
    library = clones[0].library
    CNratios = {}
    for chrname in clones[0].genome.chrom_names:
        CNratios[chrname] = []
        for a in [0, 1]:
            arm_counts = [clone.attributes.arm_counts[chrname][a] for clone in clones]
            avg_arm_count = np.average(arm_counts, weights=popsizes)
            CNrat = avg_arm_count / library.arm_sizes[chrname][a]
            CNratios[chrname].append(CNrat)
    return CNratios


import numpy as np

import string
from itertools import product
from abc import ABC, abstractmethod
from typing import Optional, Union, List

from sistem.selection import BaseLibrary
from sistem.anatomy.utils import create_random_distances
from sistem.utilities.IO import load_distances_from_file
from sistem.parameters import Parameters, fill_params

class BaseAnatomy(ABC):
    """Abstract base class for migration models. If creating a custom migration model, users must create a child class which inherits from this class. See :ref:`Parameters <parameters>` for an explanation of the constructor parameters.

    Migration are used to define a distance function :math:`d(a,b)` between each pair of sites a and b. The probability that cell :math:`c` migrates from :math:`a` to :math:`b` is given as 

    .. math::

        \\alpha_c(a, b) = \\left\\{\\begin{array}{lr} 0 & \\text{if } a = b, \\\\ \\frac{1}{d(a,b)} \\cdot \\epsilon & \\text{otherwise}\\end{array}\\right\\},

    where :math:`\\epsilon` is a per-generation baseline migration rate.

    Args:
        libraries (BaseLibrary, List[BaseLibrary]): The primary site selection library, or a list of libraries (one for each site). Required.
        params (Parameters, optional):
        nsites (int, optional):
        growth_rate (float, optional):
        max_growth_rate_multiplier (int, float, optional):
        capacities (int, List[int], optional):
        epsilon (float, optional):
        t_max (int, optional):
        N0 (int, optional):

    Attributes:
        exp_pop (List[list]): A list of lists, one for each site, containing the expected population size for each generation t=0,...,t_max.
        dists (array): The pairwise distances between anatomical sites, in order of the upper triangular indices.
        points (array): If initialized with random distances, contains the 2d coordinates of each site used to derive the distances.
    """
    def __init__(
        self, 
        libraries: Optional[Union[BaseLibrary, List[BaseLibrary]]] = None,
        params: Optional[Parameters] = None,
        *,
        nsites: Optional[int] = None, 
        growth_rate: Optional[float] = None, 
        max_growth_rate_multiplier: Optional[Union[int, float]] = None, 
        capacities: Optional[Union[int, List[int]]] = None, 
        epsilon: Optional[float] = None, 
        t_max: Optional[int] = None, 
        N0: Optional[int] = None
    ):
        params = fill_params(params, nsites=nsites, growth_rate=growth_rate, max_growth_rate_multiplier=max_growth_rate_multiplier, capacities=capacities, epsilon=epsilon, t_max=t_max, N0=N0)
        
        self.nsites = params.nsites
        self.growth_rate = params.growth_rate
        self.eps = params.epsilon
        self.t_max = params.t_max
        self.N0 = params.N0
        if params.max_growth_rate_multiplier is None or params.max_growth_rate_multiplier < 1:
            self.max_meta_growth_rate = self.growth_rate
        else:
            self.max_meta_growth_rate = self.growth_rate*params.max_growth_rate_multiplier
        self.capacities = self._check_capacities(params.capacities)
        self.exp_pop = [[] for i in range(self.nsites)]

        chars = list(string.ascii_uppercase)
        chars.remove('P')
        self.site_ids = ['P']
        if self.nsites > 26:
            chars = [''.join(x) for x in product(chars, repeat=2)]
        self.site_ids += chars[:self.nsites - 1]

        self.init_pop_dynamic(0, 0, params.t_max, N0=params.N0)

        self.libraries = self._check_libraries(libraries)
        self.dists = None
        self.points = None
        self.initialized = False

    def _check_capacities(self, capacities):
        if isinstance(capacities, list):
            if len(capacities) == 1:
                return capacities*self.nsites
            if len(capacities) != self.nsites:
                raise ValueError("Number of capacities does not match number of sites.")
            return capacities
        elif isinstance(capacities, (int, float)):
            return [int(capacities)]*self.nsites
        else:
            raise ValueError("The provided 'capacities' argument must be an int or list of ints.")
    
    def _check_libraries(self, libraries):
        if libraries is None:
            raise ValueError("Argument 'libraries' is required and cannot be None. Provide either a single Library class instance or a list of Library class instances.")
        
        if isinstance(libraries, list):
            if not all(isinstance(lib, BaseLibrary) for lib in libraries):
                raise TypeError("The provided 'libraries' list contains elements which are not Library class instances.")
            
            if len(libraries) != self.nsites:
                raise ValueError("If passing a list of libraries, ensure the number of libraries is equal to the number of sites.")
            
            primary_library_type = type(libraries[0])
            if not all(type(lib) is primary_library_type for lib in libraries):
                raise TypeError("All elements must be instances of the same Library class.")
            return libraries

        elif isinstance(libraries, BaseLibrary):
            return [libraries] * self.nsites
        
        else:
            raise TypeError("The provided 'libraries' argument must be a Library class instance or a list of Library class instances.")
    
    def init_pop_dynamic(self, site, t, t_max, N0=None, cur_fit=None, library=None):
        pops = [0 for i in range(t)]
        if N0 is not None:
            pops.append(N0)
            gr = self.growth_rate
        else:
            pops.append(self.N0)
            if self.growth_rate == self.max_meta_growth_rate or library == None:
                gr = self.growth_rate
            else:
                r = np.log(1 + cur_fit - library.base_fit) / np.log(1 + library.max_fit - library.base_fit)
                gr = self.growth_rate + r * (self.max_meta_growth_rate - self.growth_rate)
        for i in range(t, t_max+1):
            pn = pops[i]
            pn1 = pn + gr*pn*(1 - pn/self.capacities[site])
            pops.append(pn1)
        self.exp_pop[site] = pops

    def initialize_distances(self, method='random', matrix=None, path=None):
        """Method used to initialize the organotropism priors. 

        Args:
            method (str): The method used to initialize the distances. Either 'random' or 'precomputed'. If precomputed, must provide either a matrix or path. Note that if precomputed, the distance between a and itself must be 0.
            matrix (list, np.ndarray): A pairwise distance matrix. Must either be square or 1d containing just the upper triangular indices in order. 
            path (str): Instead of passing a python object, users can also provide a path to a file containing the distance matrix. Each row of the matrix should be on a separate line, and values should be separated by 0s. 
        """
        if method == 'random':
            points, norm_dists = create_random_distances(self.nsites)
            self.points = points
            self.dists = norm_dists
        elif method == 'precomputed':
            if matrix is not None:
                if not isinstance(matrix, (list, np.ndarray)):
                    raise ValueError("If providing a precomputed distance matrix, must be either a list or numpy array.")
                matrix = np.array(matrix)
                if matrix.ndim == 1:
                    self.dists = matrix
                else:
                    idxs = np.triu_indices(self.nsites)
                    self.dists = matrix[idxs]
            elif path is not None:
                self.dists = load_distances_from_file(self.nsites, path)
            else:
                raise ValueError("If using precomputed distances, either 'matrix' (list or numpy array) or 'path' (path to file containing distances) must be specified.")

        else:
            raise ValueError("Argument 'method' must be either 'random' or 'precomputed'.")
    
    def create_random_metastatic_libraries(
        self, 
        method: str = 'random',
        params: Optional[Parameters] = None,
        alter_prop: Optional[float] = None, 
        CN_coeff: Optional[float] = None
    ):
        """Creates site-specific selection libraries. The anatomy class defaults to using the same selection library for all sites, and this method is only used if site-specific selection is desired. There are two options for the *method* parameter: 'random' or 'distance'. If random, the number of selection coefficients in each metastatic site will differ from the primary site with a ratio near *alter_prop*. If distance, the number of selection coefficients that differ between all pairs of sites will roughly align with the pairwise distance matrix created with the :code:`initialize_distance` method See :ref:`Parameters <parameters>` for an explanation of the parameters.

        Args:
            method (str): The method used to create the metastatic libraries. Must be either 'random' or 'distance'.
            params (Parameters, optional):
            alter_prop (float, optional):
            CN_coeff (float, optional):
            
        """
        params = fill_params(params, alter_prop=alter_prop, CN_coeff=CN_coeff)

        if self.nsites == 1:
            raise ValueError("Number of sites must be >1 to create metastatic libraries.")
        primary_library = self.libraries[0]
        primary_library_type = type(primary_library)
        if hasattr(primary_library_type, 'create_metastatic_libraries') and callable(getattr(primary_library_type, 'create_metastatic_libraries')):
            self.libraries = primary_library.create_metastatic_libraries(self.nsites, params.alter_prop, params.CN_coeff, method=method, dists=self.dists)
        else:
            raise AttributeError(f"{primary_library_type.__name__} must have a callable 'create_metastatic_libraries' method in order to call this function.")

    
    @abstractmethod
    def get_oneway_transition_probabilities(self, clone):
        """Method which returns an array of the migation probabilities between the clone and all sites at the current generation

        Args:
            clone (Clone, optional): The clone.

        Returns:
            list: The migration probabilities.
        """
        pass

    @abstractmethod
    def get_total_transition_probability(self, clone):
        """Method which returns the total probability that a clone migrates to any other site.

        Args:
            clone (Clone, optional): The clone.

        Returns:
            float: The total migration probability.
        """
        pass

class SimpleAnatomy(BaseAnatomy):
    """The simplest migration model. Use for simulating single-site experiments, or if you want an identical fixed migration probability to each site for every cell at every generation (:math:`d(a,b)=1` for all :math:`a,b`).

    """
    def __init__(self, libraries: Optional[Union[BaseLibrary, List[BaseLibrary]]], **kwargs):
        super().__init__(libraries=libraries, **kwargs)
        self._tmatrix = np.array([[self.eps for j in range(i)] + [0] + [self.eps for j in range(i+1, self.nsites)] for i in range(self.nsites)])
        self.initialized = True

    def get_oneway_transition_probabilities(self, clone):
        return self._tmatrix[clone.site]

    def get_total_transition_probability(self, clone):
        return 1 - (1 - self.eps)**self.nsites


import numpy as np

from sistem.lineage.cell import Cell
from sistem.lineage.tree import Tree
from sistem.lineage.utils import name_cell_tree

def create_clone_tree(observed):
    """
    Creates a tree from observed clones tracing backwards in time. Root will always be the starting clone.

    Args:
        observed (defaultdict): The observed clones at the time of sampling as keys with the number of sampled cells as values.
        nsites (int): The number of anatomical sites.

    Returns:
        Tree: An unresolved clone tree.
    """
    queue = list(observed.keys())
    nodes = [clone for clone in queue]
    serviced = set()
    while queue:
        cur = queue.pop(0)
        par = cur.parent
        serviced.add(cur)
        if par != None:
            if par not in nodes:
                nodes.append(par)
            par.children.append(cur)
            if par not in serviced and par not in queue:
                queue.append(par)

    for clone in nodes:
        if clone.parent == None:
            root = clone
    t = Tree(root=root)
    #name_clone_tree(t, nsites)
    t.set_branchlengths()
    return t

def create_converted_tree(tree, observed):
    clones = [clone for clone in tree.iter_postorder()]
    converter = {}
    for clone in clones:
        n = Cell(library=clone.library, site=clone.site, birth_gen=clone.birth_gen, name=clone.name)
        n.birth_counts = clone.birth_counts
        n.events = clone.events
        converter[clone] = n
    for clone in clones:
        cell = converter[clone]
        for child in clone.children:
            cell.children.append(converter[child])
        if clone.parent is not None:
            cell.parent = converter[clone.parent]

    new_tree = Tree(root=converter[tree.root])
    new_tree.set_branchlengths()
    observed_converted = {converter[k]: v for k,v in observed.items()}
    return new_tree, observed_converted

# Creates an individual Cell object for each observed cell of a clone and adds them as children
def create_singlecell_tree(tree, observed, term_gen):
    clones = [clone for clone in tree.iter_postorder()]
    converter = {}
    for clone in clones:
        n = Cell(library=clone.library, site=clone.site, birth_gen=clone.birth_gen, name=clone.name)
        n.birth_counts = clone.birth_counts
        n.events = clone.events
        converter[clone] = n
    for clone in clones:
        cell = converter[clone]
        for child in clone.children:
            cell.children.append(converter[child])
        if clone.parent is not None:
            cell.parent = converter[clone.parent]

    new_tree = Tree(root=converter[tree.root])
    new_tree.set_branchlengths()
    observed_converted = {converter[k]: v for k,v in observed.items()}

    for cell in new_tree.iter_descendants():
        if cell in observed_converted:
            nsampled = observed_converted[cell]
            for _ in range(nsampled):
                n = Cell(library=cell.library, site=cell.site, birth_gen=term_gen)
                cell.children.append(n)
                n.parent = cell
                
    return new_tree

# Takes as input a clone tree, and creates node objects for each individual sampled cell and resolves the topology
def resolve_multifurcation(tree, name_tree=True):
    nodes = [node for node in tree.iter_descendants() if not node.is_leaf()]
    while nodes:
        clone = nodes.pop(0)
        if len(clone.children) > 1:
            resolve_multifurcation_node(clone)
    if name_tree:
        name_cell_tree(tree)
    tree.set_branchlengths()

# Function which resolves multifurcating nodes by randomly selecting replication groups from the larger population. 
def resolve_multifurcation_node(clone, verbose=False):
    children = [c for c in clone.children]
    assert len(children) > 1, "Cannot resolve multifurcation with fewer than 2 children"

    groups = []
    t0 = clone.birth_gen
    t = max([clone.birth_gen for clone in children])

    if verbose:
        print(f'Resolving {clone.name}, {clone.birth_gen} birth_gen, {len(clone.children)} children')
    #diverge_count = {tt: 0 for tt in clone.birth_counts.keys()}
    #for child in children:
    #    diverge_count[child.birth_gen] += 1

    while t > t0:
        # Draw cell from greater pop for existing replication groups and add nodes if coalescence occurs
        # Add any children born in gen t to the set of replication groups
        for i in list(range(len(children)))[::-1]:
            if children[i].birth_gen == t:
                groups.append(children[i])
                children.pop(i)

        if verbose:
            print(f'{t}: {clone.birth_counts[t]}, {len(groups)}')
                
        assert len(groups) <= 2*clone.birth_counts[t], 'Unable to coalesce: not enough living cells from previous generation'
        pool = np.repeat(np.arange(clone.birth_counts[t]), 2)
        np.random.shuffle(pool)
        #pool = pool[diverge_count[t]:]
        assignments = pool[:len(groups)]
        draws = {}
        new_groups, remove_idxs = [], set()
        for i,b in enumerate(assignments):
            if b in draws:
                j = draws[b]
                n = Cell(library=clone.library, site=clone.site, birth_gen=t)
                n.children = [groups[i], groups[j]]
                groups[i].parent = groups[j].parent = n
                new_groups.append(n)
                remove_idxs.update((i,j))
            else:
                draws[b] = i
        
        groups = [x for i,x in enumerate(groups) if i not in remove_idxs] + new_groups

        t -= 1
    
    # Connect parent to new children in groups. If only one remaining group (creating unifurcation), leave as is for now.
    while len(groups) > 2:
        np.random.shuffle(groups)
        g1 = groups.pop(0)
        g2 = groups.pop(0)
        n = Cell(library=clone.library, site=clone.site, birth_gen=t)
        n.children = [g1, g2]
        g1.parent = g2.parent = n
        groups.append(n)

    clone.children = []
    for g in groups:
        clone.children.append(g)
        g.parent = clone
    
    if verbose:
        print('')

def merge_and_resolve_clone_tree(clone_tree, observed, nsites, term_gen, min_mut_fraction):
    """
    
    """

    # First add one child cell to each observed clone, then resolve multifurcations
    sample_counts = {}
    for clone, count in observed.items():
        n = Cell(library=clone.library, site=clone.site, birth_gen=term_gen)
        clone.add_child(n)
        sample_counts[n] = count
    resolve_multifurcation(clone_tree)

    # For each node, count the total number of samples cells in each site which are descendant from that node
    subtree_counts = {node: np.zeros(nsites) for node in clone_tree.traverse()}
    for node in clone_tree.iter_postorder():
        if node.is_leaf():
            subtree_counts[node][node.site] += sample_counts[node]
        else:
            for child in node.children:
                subtree_counts[node] += subtree_counts[child]

    # Helper function which, given a node and a site, removes all descendant clones (possibly itself) which belong to the site
    def remove_subclones_in_site(node, site, exclude=[]):
        if node.is_leaf():
            if node.site == site:
                if node not in exclude:
                    del sample_counts[node]
                    node.detach()
        else:
            children = node.children[:]
            for child in children:
                remove_subclones_in_site(child, site, exclude=exclude)
            if len(node.children) == 0:
                node.detach()
    
    # Helper function which, for each site, looks for subtrees in which the total number of observed cells from that site is less than some minimum threshold (min_cell_per_site). If so, will either prune the subtree or merge it with other clones.
    # min_cell_per_site: 
    def merge_clones(node, site, min_cell_per_site):
        if node.is_leaf():
            # If at this point, the leaf should already exceed the threshold, otherwise it would've been removed.
            return
        if subtree_counts[node][site] >= min_cell_per_site:
            if len(node.children) == 1:
                merge_clones(node.children[0], site, min_cell_per_site)
            else:
                assert len(node.children) == 2, 'Somehow more than 2 children'
                left, right = node.children[0], node.children[1]
                # case 1: both subtrees exceeds the lower bound
                if subtree_counts[left][site] >= min_cell_per_site and subtree_counts[right][site] >= min_cell_per_site:
                    merge_clones(left, site, min_cell_per_site)
                    merge_clones(right, site, min_cell_per_site)
                # case 2: only the left subtree passes, prune the right
                elif subtree_counts[left][site] >= min_cell_per_site:
                    remove_subclones_in_site(right, site)
                    merge_clones(left, site, min_cell_per_site)
                # case 3: only the right subtree passes, prune the left
                elif subtree_counts[right][site] >= min_cell_per_site:
                    remove_subclones_in_site(left, site)
                    merge_clones(right, site, min_cell_per_site)
                # case 4: neither passes
                else:
                    # Need to combine in a single clone. For now, we'll just pretend all of the other clones belong to 'largest'.
                    assert subtree_counts[left][site] + subtree_counts[right][site] >= min_cell_per_site, 'This should be the only remaining case'
                    leaves = [leaf for leaf in node.iter_leaves() if leaf.site == site]
                    leaves.sort(key = lambda x: sample_counts[x], reverse=True)
                    largest = leaves[0]
                    sample_counts[largest] = sum([sample_counts[leaf] for leaf in leaves])
                    #for leaf in leaves[1:]:
                    #    del sample_counts[leaf]
                    remove_subclones_in_site(node, site, exclude=[largest])

        # No possible way to overcome threshold, so just remove all the clones in the site in the subtree
        else:
            remove_subclones_in_site(node, site)

    # Counts the total number of observed cells in each site (equal to [ncells_prim, ncells_meta, ncells_meta, ...])
    tot_observed = [0 for _ in range(nsites)]
    for leaf,count in sample_counts.items():
        tot_observed[leaf.site] += count

    # Compute min_cell_per_site as the total number of observed cells in that site * a minimum mutation fraction.
    for s in range(nsites):
        min_cell_per_site = int(tot_observed[s]*min_mut_fraction)
        merge_clones(node, s, min_cell_per_site)

    #add_dummy_diploid(clone_tree, term_gen)
    #collapse_tree(clone_tree)
    #name_clone_tree(clone_tree, nsites)
    return sample_counts

def collapse_tree(tree):
    def collapse_tree_helper(node):
        if len(node.children) == 1:
            child = node.children[0]
            child.events = node.events + child.events
            #if not child.is_leaf():
            #    child.site = node.site
            child.detach()
            if node.parent == None:
                tree.root = child
            else:
                parent = node.parent
                node.detach()
                parent.add_child(child)
            collapse_tree_helper(child)
        elif len(node.children) > 1:
            children = node.children[:]
            for child in children:
                collapse_tree_helper(child)
    
    collapse_tree_helper(tree.root)
    tree.set_branchlengths()


import numpy as np
from scipy.stats import beta
from scipy.optimize import newton_krylov, NoConvergence

from sistem.utilities.utilities import get_reg_id

# Estimating alpha/beta from point on lorenz curve
def get_alpha_beta(x0, y0):
    def F(P):
        Aa, Bb = P[0], P[1]
        X = beta.cdf(float(Aa)/(Aa+Bb), Aa, Bb) - x0
        Y = beta.cdf(float(Aa)/(Aa+Bb), Aa+1, Bb) - y0
        return [X, Y]
    guess = [10, 5]
    max_n = 1000
    n = 1
    while n < max_n:
        try:
            sol = newton_krylov(F, guess, method = 'lgmres', verbose = 0, rdiff = 0.1, maxiter=50)
            break
        except NoConvergence as e:
            guess = np.random.rand(2) * 10 + 0.1
        except ValueError as e:
            guess = np.random.rand(2) * 10 + 0.1
        n += 1
    if n == max_n:
        # Error has occurred here
        return [1.38, 1.38]
    else:
        return sol

def bezier_coef(points):
    n = len(points) - 1

    C = 4 * np.identity(n)
    np.fill_diagonal(C[1:], 1)
    np.fill_diagonal(C[:, 1:], 1)
    C[0, 0] = 2
    C[n - 1, n - 1] = 7
    C[n - 1, n - 2] = 2

    P = [2 * (2 * points[i] + points[i + 1]) for i in range(n)]
    P[0] = points[0] + 2 * points[1]
    P[n - 1] = 8 * points[n - 1] + points[n]

    A = np.linalg.solve(C, P)
    B = [0] * n
    for i in range(n - 1):
        B[i] = 2 * points[i + 1] - A[i + 1]
    B[n - 1] = (A[n - 1] + points[n]) / 2

    return A, B

def single_cubic_bezier(a, b, c, d):
    return lambda t: np.power(1 - t, 3) * a + 3 * np.power(1 - t, 2) * t * b + 3 * (1 - t) * np.power(t, 2) * c + np.power(t, 3) * d

def multiple_cubic_bezier(points):
    A, B = bezier_coef(points)
    return [single_cubic_bezier(points[i], A[i], B[i], points[i + 1]) for i in range(len(points) - 1)]
    
# Generates starting coverage points at set intervals 
def gen_start_interval(num_windows, interval, Aa, Bb):
    # Sample point every interval bin
    x = [i for i in range(0, num_windows, interval)]
    if x[-1] != num_windows-1:
        if num_windows - x[-1] >= interval/2:
            x.append(num_windows-1)
        else:
            x[-1] = num_windows-1
    # draw coverage from beta distribution
    y = list(np.random.beta(Aa, Bb, len(x)))
    points = [np.array(p) for p in zip(x, y)]
    return points

# Smooths points across bins with bezier curves
def gen_coverage(num_windows, interval, Aa, Bb):
    points = gen_start_interval(num_windows, interval, Aa, Bb)
    A, B = bezier_coef(points)
    curves = [single_cubic_bezier(points[i], A[i], B[i], points[i + 1]) for i in range(len(points) - 1)]

    new_points = []
    for i in range(len(points) - 1):
        f = curves[i]
        gaps = points[i+1][0] - points[i][0] + 1
        coords = [f(t) for t in np.linspace(0, 1, int(gaps))]
        if i == 0:
            new_points.append((round(coords[0][0]), coords[0][1]))
        new_points += [(round(i[0]), i[1]) for i in coords[1:]]
    
    new_points.sort(key=lambda x: x[0])
    # Multiply by 2 so that values are in range 0-2. Value of 1 indicates mean coverage.
    new_points = [2*max(min(i[1], 1), 0) for i in new_points]
    return new_points

def bin_region_values(vals, r2b):
    """
    Given an array of values and the number of regions in each bin, returns a new array containing the binwise averages. Allows the ratio to be non-integral.

    Arguments:
        vals (array[int]): Value of arrays (e.g. Copy numbers, readcounts.)
        r2b (int, float): The number of regions per bin (e.g. 2, 2.5).
    """
    res = []
    tot_sum = 0
    n = len(vals)
    while tot_sum < n:
        idx = int(tot_sum)

        left_w = np.ceil(tot_sum) - tot_sum
        if left_w == 0:
            left_w = 1
        
        binned_vals = [vals[idx]]
        weights = [left_w]

        if tot_sum + r2b > n:
            for i in range(idx + 1, n):
                binned_vals.append(vals[i])
                weights.append(1)
            
        else:
            full_w = int(r2b - left_w)
            right_w = (r2b - left_w) % 1

            binned_vals.extend([vals[idx + i + 1] for i in range(full_w)])
            
            weights.extend([1 for i in range(full_w)])
            if right_w > 0:
                binned_vals.append(vals[idx + full_w + 1])
                weights.append(right_w)

        res.append(float(np.average(binned_vals, weights=weights)))
        tot_sum += r2b

    return res

def bin_region_readcounts(vals, r2b):
    """
    Given an array of readcounts and the number of regions in each bin, returns a new array containing the binwise sums. Allows the ratio to be non-integral. Normalizez for the number of regions.

    Arguments:
        vals (array[int]): Value of arrays (e.g. Copy numbers, readcounts.)
        r2b (int, float): The number of regions per bin (e.g. 2, 2.5).
    """
    res = []
    tot_sum = 0
    n = len(vals)
    while tot_sum < n:
        idx = int(tot_sum)

        left_w = np.ceil(tot_sum) - tot_sum
        if left_w == 0:
            left_w = 1
        
        binned_vals = [vals[idx]]
        weights = [left_w]

        if tot_sum + r2b > n:
            for i in range(idx + 1, n):
                binned_vals.append(vals[i])
                weights.append(1)
            
        else:
            full_w = int(r2b - left_w)
            right_w = (r2b - left_w) % 1

            binned_vals.extend([vals[idx + i + 1] for i in range(full_w)])
            
            weights.extend([1 for i in range(full_w)])
            if right_w > 0:
                binned_vals.append(vals[idx + full_w + 1])
                weights.append(right_w)

        next_val = 0
        for i,x in enumerate(binned_vals):
            next_val += x*weights[i]
        res.append(round(next_val * (min(r2b, n - tot_sum)) / r2b))
        tot_sum += r2b

    return res

def count_region_CNs(cell, total=False):
    profile = {}
    for chrname, nregions in cell.library.regions.items():
        CNs = [np.zeros(nregions, dtype=int), np.zeros(nregions, dtype=int)]
        for chromosome in cell.genome[chrname]:
            counts = np.bincount([get_reg_id(x) for x in chromosome.seq], minlength=nregions)
            CNs[chromosome.allele] += counts
        
        if total:
            profile[chrname] = CNs[0] + CNs[1]
        else:
            profile[chrname] = CNs
    return profile


def get_mutated_basepairs(observed_cells, chrnames, region_len):
    mutated_bps = {chrname: set() for chrname in chrnames}
    for chrname in chrnames:
        for cell in observed_cells:
            for chromosome in cell.genome[chrname]:
                for r, positions in chromosome.SNVs.items():
                    v = get_reg_id(int(r))
                    for p in positions:
                        loc = int(v*region_len + p)
                        mutated_bps[chrname].add(loc)
        mutated_bps[chrname] = list(mutated_bps[chrname])
        mutated_bps[chrname].sort()
    mutated_bps = [(chrname, pos) for chrname in chrnames for pos in mutated_bps[chrname]]
    return mutated_bps